import dataclasses as dataclasses
import abc
from typing import Callable, Iterable, List, Literal, Optional, Tuple, Union
import numpy as np
import math
import scipy
import moval.models
import moval.solvers
from moval.solvers import register

@register("base-solver")
@dataclasses.dataclass
class Solver(abc.ABC):
    """Solver base class.

    A solver contains helper methods for bundling a model, criterion and optimizer.
    This solver will save the results in :py:attr:`models.param`.

    Atrributes:
        model (moval.models.Model): 
            The model to calibrate the network outputs.
        metric (str):
            The considered performance metric to align.
        class_specific (bool):
            If ``True``, the calculation will match class-wise confidence to class-wise accuracy/DSC.
        criterions (moval.solver.Calibrate):
            Calibration criterions. This will be utilized by solvers.

    """

    model: moval.models.Model
    metric: str
    class_specific: bool = dataclasses.field(init=False)
    
    def __post_init__(self):
        self.class_specific = self.model.class_specific

        if self.model.mode == "classification":
            self.criterions = moval.solvers.clsCalibrate(class_specific = self.class_specific, metric = self.metric)
        elif self.model.mode == "segmentation":
            self.criterions = moval.solvers.segCalibrate(class_specific = self.class_specific, metric = self.metric)
        else:
            raise ValueError(f"Unknown mode '{self.mode}'")

    def eval_func(self, x: float) -> float:
        """The evaluation function for scipy optimization.
        
        Args:
            x: The variable (numpy.float64) to be optimized.
        
        Return:
            err: The caculated err taking x as input.

        """

        if not self.class_specific:
            self.model.param = x
            if self.model.extend_param:
                estim_acc, _ = self.model.estimate_accuracy(self.inp, midstage = True, gt_guide = self.gt_guide)
            else:
                estim_acc, _ = self.model.estimate_accuracy(self.inp, gt_guide = self.gt_guide)
            err = self.criterions(self.inp, self.gt, estim_acc)
        else:
            estim_perf_all = []
            kcls_all = []
            for kcls in range(self.batch * self.opt_kcls, np.min((self.batch * (self.opt_kcls + 1), self.model.num_class))):

                self.model.param[kcls] = x
                
                if self.metric == "accuracy":
                    if self.model.extend_param:
                        _, estim_perf = self.model.estimate_accuracy(self.inp, midstage = True, gt_guide = self.gt_guide)
                    else:
                        _, estim_perf = self.model.estimate_accuracy(self.inp, gt_guide = self.gt_guide)
                    estim_perf_all.append(estim_perf[kcls])
                elif self.metric == "sensitivity":
                    probability = self.model.calculate_probability(self.inp, midstage = True, appr=True)
                    estim_perf = self.model.estimate_sensitivity(self.inp, probability, gt_guide = self.gt_guide)
                    estim_perf_all.append(estim_perf[kcls])
                elif self.metric == "precision":
                    probability = self.model.calculate_probability(self.inp, midstage = True, appr=True, full=True)
                    estim_perf = self.model.estimate_precision(probability, gt_guide = self.gt_guide)
                    estim_perf_all.append(estim_perf[kcls])
                elif self.metric == "f1score":
                    probability = self.model.calculate_probability(self.inp, midstage = True, appr=True)
                    estim_perf = self.model.estimate_f1score(self.inp, probability, gt_guide = self.gt_guide)
                    estim_perf_all.append(estim_perf[kcls])
                elif self.metric == "auc":
                    probability = self.model.calculate_probability(self.inp, midstage = True, appr=True, full=True)
                    estim_perf = self.model.estimate_auc(probability, gt_guide = self.gt_guide, sel_cls = kcls)
                    estim_perf_all.append(estim_perf[0])
                else:
                    ValueError(f"Unsupported metric '{self.metric}'")
                kcls_all.append(kcls)

            err = self.criterions(self.inp, self.gt, estim_perf_all, kcls_all)

        return err

    def eval_func_ext(self, x: float) -> float:
        """The evaluation function for scipy optimization of extended parameters.
        
        Args:
            x: The variable to be optimized.
        
        Return:
            err: The caculated err taking x as input.

        """
        if not self.class_specific:
            self.model.param_ext = x
            estim_acc, _ = self.model.estimate_accuracy(self.inp, gt_guide = self.gt_guide)
            err = self.criterions(self.inp, self.gt, estim_acc)
        else:
            estim_perf_all = []
            kcls_all = []
            for kcls in range(self.batch * self.opt_kcls, np.min((self.batch * (self.opt_kcls + 1), self.model.num_class))):
                self.model.param_ext[kcls] = x

                if self.metric == "accuracy":
                    _, estim_perf = self.model.estimate_accuracy(self.inp, gt_guide = self.gt_guide)
                    estim_perf_all.append(estim_perf[kcls])
                elif self.metric == "sensitivity":
                    probability = self.model.calculate_probability(self.inp, appr=True)
                    estim_perf = self.model.estimate_sensitivity(self.inp, probability, gt_guide = self.gt_guide)
                    estim_perf_all.append(estim_perf[kcls])
                elif self.metric == "precision":
                    probability = self.model.calculate_probability(self.inp, appr=True, full=True)
                    estim_perf = self.model.estimate_precision(probability, gt_guide = self.gt_guide)
                    estim_perf_all.append(estim_perf[kcls])
                elif self.metric == "f1score":
                    probability = self.model.calculate_probability(self.inp, appr=True)
                    estim_perf = self.model.estimate_f1score(self.inp, probability, gt_guide = self.gt_guide)
                    estim_perf_all.append(estim_perf[kcls])
                elif self.metric == "auc":
                    probability = self.model.calculate_probability(self.inp, appr=True, full=True)
                    estim_perf = self.model.estimate_auc(probability, gt_guide = self.gt_guide, sel_cls = self.kcls)
                    estim_perf_all.append(estim_perf[0])
                else:
                    ValueError(f"Unsupported metric '{self.metric}'")
                kcls_all.append(kcls)

            err = self.criterions(self.inp, self.gt, estim_perf_all, kcls_all)

        return err

    def kcls_order_list(self, inp: np.ndarray, exclusive_background: bool = False) -> List[Iterable]:
        """Generate a list of kcls, such that the predicted samples are in descending order.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` for classification and a list of n ``(d, H, W, (D))`` for segmentation. 
            exclusive_background: If ``False``, we exclude the background class (the most majority class).
        
        Return:
            kcls_list: a list of class index. The first element should correspond to the most major class.

        """
        kcls_sample = np.zeros(self.opt_class)
        if self.model.mode == "classification":
            pred = np.argmax(inp, axis = 1)
            for opt_kcls in range(self.opt_class):
                for kcls in range(self.batch * opt_kcls, np.min((self.batch * (opt_kcls + 1), self.model.num_class))):
                    kcls_sample[opt_kcls] = kcls_sample[opt_kcls] + np.sum(pred == kcls)
        else:
            for kcls in range(self.model.num_class):
                for n_case in range(len(inp)):
                    pred_case = np.argmax(inp[n_case], axis = 0)
                    kcls_sample[kcls] = kcls_sample[kcls] + np.sum(pred_case == kcls)
        
        kcls_list = np.argsort(kcls_sample)
        if exclusive_background:
            kcls_list = kcls_list[:-1]

        return kcls_list[::-1]

    def fit(
        self,
        inp: Union[List[Iterable], np.ndarray],
        gt: Union[List[Iterable], np.ndarray],
        batch: int = 1
    ) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
        """Fit the model based on scipy backend.
        
        Args:
            inp: The network output (logits) of shape ``(n, d)`` for classification and a list of n ``(d, H, W, (D))`` for segmentation. 
            gt: The cooresponding annotation of shape ``(n, )`` for classification and a list of n ``(H, W, (D))`` for segmentation.
            batch: To match the group-wise accuracy with group-wise confidence score. batch is the group size. This is useful when validation data per case is few.
        
        Note:
            ``batch`` is only used for class-specific variants. It does NOT support segmentation tasks as it is always unnecessary.
            
        Return:
            param: The optimized parameters, of shape ``(d, )`` or a float.
            If the model contains extended parameters, will return Tuple[param, param_ext].

        """

        print(f"Starting optimizing for model {self.model.estim_algorithm} with confidence {self.model.confidence_scores} based on metric {self.metric}, class specific is {self.model.class_specific}.")

        if self.model.estim_algorithm == "ac-model":
            # do not need optimization, end task as soon as we possible!

            self.model.train()

            if self.model.conf.normalization:
                _, _ = self.model.estimate_accuracy(inp)

            self.model.eval()
            self.model.is_fitted = True

            return self.model.param

        self.inp = inp
        self.gt = gt
        self.batch = batch
        self.opt_class = math.ceil(self.model.num_class / batch)
        # generate gt_guide from gt here.
        if self.model.mode == "segmentation":
            gt_guide = []
            for n_case in range(len(gt)):
                gt_case     = gt[n_case]
                gt_exist = []
                for k_cls in range(self.model.num_class):
                    gt_exist.append(np.sum(gt_case == k_cls) > 0)
                gt_guide.append(gt_exist)
            gt_guide = np.array(gt_guide)
            self.gt_guide = gt_guide
        else:
            self.gt_guide = None

        self.model.train()
        
        # some hyper parameters.
        optimization_method = 'Nelder-Mead'
        x0 = np.array([1.0])
        search_threshold = 0.01 # if the optimized results are larger than this, we go through the initial conditions
        exclusive_background = False

        if self.model.mode == "classification":
            if self.model.estim_algorithm == "doc-model":
                # the range of doc estimation is [0, 2].
                initial_conditions = [np.array([0.01]), np.array([0.1]), np.array([0.5]), np.array([1.5]), np.array([1.8])]
            else:
                # the range of ts estimation is [0, inf].
                initial_conditions = [np.array([0.1]), np.array([0.5]), np.array([3.]), np.array([5.]), np.array([10.])]
            print(f"Opitimizing with {inp.shape[0]} samples...")
            
        else:
            if self.model.estim_algorithm == "doc-model":
                # the range of doc estimation is [0, 2].
                initial_conditions = [np.array([0.01]), np.array([1.5])]
            else:
                # the range of ts estimation is [0, inf].
                initial_conditions = [np.array([0.1]), np.array([5])]

            print(f"Opitimizing with {len(inp)} samples...")
            print("Be patient, it should take a while...")
            if self.metric != "accuracy":
                exclusive_background = True # leave the background class uncalibrated.

        # generate a list of length kcls, from high to low
        kcls_list = self.kcls_order_list(self.inp, exclusive_background)

        if self.class_specific:
            
            if self.model.mode == "classification":
                pred = np.argmax(inp, axis = 1)
            else:
                preds = []
                for n_case in range(len(inp)):
                    preds.append(np.argmax(inp[n_case], axis = 0))

            for opt_kcls in kcls_list:
                self.opt_kcls = opt_kcls

                # do the class-wise optimization process, if 
                # a) there is any corresponding predictions, as it is possible for alignment, metric = 0.
                # and
                # b) there is any corresponding GT, as the alignment would has some meaning.
                flag_case_a = False
                flag_case_b = False
                if self.model.mode == "classification":
                    for kcls in range(self.batch * opt_kcls, np.min((self.batch * (opt_kcls + 1), self.model.num_class))):
                        if np.sum(pred == kcls) > 0:
                            flag_case_a = True
                        if np.sum(gt == kcls) > 0:
                            flag_case_b = True
                else:
                    for n_case in range(len(inp)):
                        if np.sum(preds[n_case] == opt_kcls) > 0:
                            flag_case_a = True
                        if np.sum(gt[n_case] == opt_kcls) > 0:
                            flag_case_b = True

                if flag_case_a and flag_case_b:

                    optimization_result = scipy.optimize.minimize(
                                    fun = self.eval_func,
                                    x0 = x0,
                                    method = optimization_method,
                                    bounds = [(1e-06,None)],
                                    tol = 1e-07)
                    
                    optimized_param = optimization_result.x[0]

                    if optimization_result.fun > search_threshold:
                        # change the initial state, if we are not satisfied with the optimization results.
                        print(f"Not satisfied with initial optimization results of param for class {opt_kcls}, trying more initial states...")
                        if self.model.estim_algorithm == "atc-model":
                            score = self.model.conf(inp)
                            if self.model.conf.normalization:
                                score = self.model._normalize(score)
                            score_ = []
                            if self.model.mode == "classification":
                                for kcls in range(self.batch * opt_kcls, np.min((self.batch * (opt_kcls + 1), self.model.num_class))):
                                    score_.append(score[np.argmax(inp, axis = 1) == kcls])
                                score_ = np.concatenate(score_)
                                initial_conditions = [np.array([np.percentile(score_, 20)]), 
                                                      np.array([np.percentile(score_, 40)]), 
                                                      np.array([np.percentile(score_, 50)]), 
                                                      np.array([np.percentile(score_, 60)]), 
                                                      np.array([np.percentile(score_, 80)])]
                            else:
                                for n_case in range(len(inp)):
                                    score_flatten = score[n_case].flatten() # ``n``
                                    pred_flatten = np.argmax(inp[n_case], axis = 0).flatten()
                                    score_.append(score_flatten[pred_flatten == opt_kcls])
                                score_ = np.concatenate(score_)
                                initial_conditions = [np.array([np.percentile(score_, 20)]), 
                                                      np.array([np.percentile(score_, 50)])]
                        results = []
                        results.append((optimization_result.fun, optimization_result.x[0]))
                        cnt_guess = 0
                        for initial_guess in initial_conditions:
                            optimization_result = scipy.optimize.minimize(
                                    fun = self.eval_func,
                                    x0 = initial_guess,
                                    method = optimization_method,
                                    bounds = [(1e-06,None)],
                                    tol = 1e-07)
                            results.append((optimization_result.fun, optimization_result.x[0]))
                            cnt_guess += 1
                            print(f"Tried {cnt_guess}/{len(initial_conditions)} times.")
                        
                        print(f"Starting from {initial_conditions}")
                        print(f"Optimization results are {results}")

                        optimized_param = min(results, key=lambda x: x[0])[1]

                    for kcls in range(self.batch * opt_kcls, np.min((self.batch * (opt_kcls + 1), self.model.num_class))):
                        self.model.param[kcls] = optimized_param
        else:
            optimization_result = scipy.optimize.minimize(
                            fun = self.eval_func,
                            x0 = x0,
                            method = optimization_method,
                            bounds = [(1e-06,None)],
                            tol = 1e-07)
        
            optimized_param = optimization_result.x

            if optimization_result.fun > search_threshold:
                # change the initial state, if we are not satisfied with the optimization results.
                print(f"Not satisfied with initial optimization results of param, trying more initial states...")
                if self.model.estim_algorithm == "atc-model":
                    score = self.model.conf(inp)
                    if self.model.conf.normalization:
                        score = self.model._normalize(score)
                    if self.model.mode == "classification":
                        score_ = score
                        initial_conditions = [np.array([np.percentile(score_, 20)]), 
                                                    np.array([np.percentile(score_, 40)]), 
                                                    np.array([np.percentile(score_, 50)]), 
                                                    np.array([np.percentile(score_, 60)]), 
                                                    np.array([np.percentile(score_, 80)])]
                    else:
                        score_ = []
                        for n_case in range(len(inp)):
                            score_flatten = score[n_case].flatten() # ``n``
                            score_.append(score_flatten)
                        score_ = np.concatenate(score_)
                        initial_conditions = [np.array([np.percentile(score_, 20)]), 
                                                    np.array([np.percentile(score_, 50)])]
                results = []
                results.append((optimization_result.fun, optimization_result.x))
                cnt_guess = 0
                for initial_guess in initial_conditions:
                    optimization_result = scipy.optimize.minimize(
                            fun = self.eval_func,
                            x0 = initial_guess,
                            method = optimization_method,
                            bounds = [(1e-06,None)],
                            tol = 1e-07)
                    results.append((optimization_result.fun, optimization_result.x))
                    cnt_guess += 1
                    print(f"Tried {cnt_guess}/{len(initial_conditions)} times.")
                
                print(f"Starting from {initial_conditions}")
                print(f"Optimization results are {results}")

                optimized_param = min(results, key=lambda x: x[0])[1]

            self.model.param = optimized_param

        if self.model.extend_param:

            if self.class_specific:
                for opt_kcls in kcls_list:
                    self.opt_kcls = opt_kcls

                    # do the class-wise optimization process, if 
                    # a) there is any corresponding predictions, as it is possible for alignment, metric = 0.
                    # and
                    # b) there is any corresponding GT, as the alignment would has some meaning.
                    flag_case_a = False
                    flag_case_b = False
                    if self.model.mode == "classification":
                        for kcls in range(self.batch * opt_kcls, np.min((self.batch * (opt_kcls + 1), self.model.num_class))):
                            if np.sum(pred == kcls) > 0:
                                flag_case_a = True
                            if np.sum(gt == kcls) > 0:
                                flag_case_b = True
                    else:
                        for n_case in range(len(inp)):
                            if np.sum(preds[n_case] == opt_kcls) > 0:
                                flag_case_a = True
                            if np.sum(gt[n_case] == opt_kcls) > 0:
                                flag_case_b = True

                    if flag_case_a and flag_case_b:

                        optimization_result = scipy.optimize.minimize(
                                        fun = self.eval_func_ext,
                                        x0 = x0,
                                        method = optimization_method,
                                        bounds = [(1e-03,None)],
                                        tol = 1e-07)
                        
                        optimized_param_ext = optimization_result.x[0]

                        if optimization_result.fun > search_threshold:
                            # change the initial state, if we are not satisfied with the optimization results.
                            print(f"Not satisfied with initial optimization results of param_ext for class {opt_kcls}, trying more initial states...")
                            # change atc to be consistent with the range of confidence score
                            # get the max and min value here.
                            score = self.model.calibrate(inp, midstage = True)
                            score_ = []
                            if self.model.mode == "classification":
                                for kcls in range(self.batch * opt_kcls, np.min((self.batch * (opt_kcls + 1), self.model.num_class))):
                                    score_.append(score[np.argmax(inp, axis = 1) == kcls])
                                score_ = np.concatenate(score_)
                                initial_conditions_atc = [np.array([np.percentile(score_, 20)]), 
                                                            np.array([np.percentile(score_, 40)]), 
                                                            np.array([np.percentile(score_, 50)]), 
                                                            np.array([np.percentile(score_, 60)]), 
                                                            np.array([np.percentile(score_, 80)])]
                            else:
                                for n_case in range(len(inp)):
                                    score_flatten = score[n_case].flatten() # ``n``
                                    pred_flatten = np.argmax(inp[n_case], axis = 0).flatten()
                                    score_.append(score_flatten[pred_flatten == opt_kcls])
                                score_ = np.concatenate(score_)
                                initial_conditions_atc = [np.array([np.percentile(score_, 20)]), 
                                                            np.array([np.percentile(score_, 50)])]
                            results = []
                            results.append((optimization_result.fun, optimization_result.x[0]))
                            cnt_guess = 0
                            for initial_guess in initial_conditions_atc:
                                optimization_result = scipy.optimize.minimize(
                                        fun = self.eval_func_ext,
                                        x0 = initial_guess,
                                        method = optimization_method,
                                        bounds = [(1e-03,None)],
                                        tol = 1e-07)
                                results.append((optimization_result.fun, optimization_result.x[0]))
                                cnt_guess += 1
                                print(f"Tried {cnt_guess}/{len(initial_conditions_atc)} times.")
                            
                            print(f"Starting from {initial_conditions_atc}")
                            print(f"Optimization results are {results}")

                            optimized_param_ext = min(results, key=lambda x: x[0])[1]

                        for kcls in range(self.batch * opt_kcls, np.min((self.batch * (opt_kcls + 1), self.model.num_class))):
                            self.model.param_ext[kcls] = optimized_param_ext

            else:
                optimization_result = scipy.optimize.minimize(
                                fun = self.eval_func_ext,
                                x0 = x0,
                                method = optimization_method,
                                bounds = [(1e-03,None)],
                                tol = 1e-07)

                optimized_param_ext = optimization_result.x

                if optimization_result.fun > search_threshold:
                    # change the initial state, if we are not satisfied with the optimization results.
                    print(f"Not satisfied with initial optimization results of param_ext, trying more initial states...")
                    # change atc to be consistent with the range of confidence score
                    # get the max and min value here.
                    score = self.model.calibrate(inp, midstage = True)
                    if self.model.mode == "classification":
                        score_ = score
                        initial_conditions_atc = [np.array([np.percentile(score_, 20)]), 
                                                    np.array([np.percentile(score_, 40)]), 
                                                    np.array([np.percentile(score_, 50)]), 
                                                    np.array([np.percentile(score_, 60)]), 
                                                    np.array([np.percentile(score_, 80)])]
                    else:
                        score_ = []
                        for n_case in range(len(inp)):
                            score_flatten = score[n_case].flatten() # ``n``
                            score_.append(score_flatten)
                        score_ = np.concatenate(score_)
                        initial_conditions_atc = [np.array([np.percentile(score_, 20)]), 
                                                    np.array([np.percentile(score_, 50)])]
                    results = []
                    results.append((optimization_result.fun, optimization_result.x))
                    cnt_guess = 0
                    for initial_guess in initial_conditions_atc:
                        optimization_result = scipy.optimize.minimize(
                                fun = self.eval_func_ext,
                                x0 = initial_guess,
                                method = optimization_method,
                                bounds = [(1e-03,None)],
                                tol = 1e-07)
                        results.append((optimization_result.fun, optimization_result.x))
                        cnt_guess += 1
                        print(f"Tried {cnt_guess}/{len(initial_conditions_atc)} times.")
                    
                    print(f"Starting from {initial_conditions_atc}")
                    print(f"Optimization results are {results}")

                    optimized_param_ext = min(results, key=lambda x: x[0])[1]

                self.model.param_ext = optimized_param_ext

        # I need save the normalization parameter, if the confidence needs to be normalized.
        _ = self.model.calibrate(self.inp)

        self.model.eval()
        self.model.is_fitted = True

        if self.model.extend_param:
            return self.model.param, self.model.param_ext
        else:
            return self.model.param



from typing import Callable, Iterable, List, Literal, Optional, Tuple, Union

import numpy as np

def cal_softmax(x: np.ndarray, T = 1, e1 = 1e-6) -> np.ndarray:
    """Compute softmax values for each sets of scores in x.
    
    Args:
        x: The network output (logits) of shape ``(n, d)``.
        T: A scarlar temperature to calibrate the score.
        e1: A small number to prevent unexpected results of division.
    
    Returns:
        prob: The calculated softmax probability for different classes of shape ``(n, d)``

    """
    # to avoid exponent exploding
    x_max = np.max(x, keepdims=True, axis = 1).repeat(x.shape[1], axis=1)

    prob = np.exp((x - x_max).transpose() / T) / (np.sum(np.exp((x - x_max).transpose() / T), axis=0) + e1)

    return prob.transpose()

def cal_energy(x: np.ndarray, T = 1) -> np.ndarray:
    """Compute energy values for each sets of scores in x.

    Calculating based on the defination as described in:
    ``https://proceedings.neurips.cc/paper/2020/hash/f5496252609c43eb8a3d147ab9b9c006-Abstract.html``
    
    Args:
        x: The network output (logits) of shape ``(n, d)``.
        T: A scarlar temperature to calibrate the score.
    
    Returns:
        energy: The calculated energy of shape ``(n, )``

    """

    denominator = np.sum(np.exp(x.transpose() / T), axis=0)
    energy = - T * np.log(denominator)

    return energy

def cal_mcp(x: np.ndarray, T = 1) -> np.ndarray:
    """Compute maximum class probability (MCP) for each sets of scores in x.
    
    Args:
        x: The network output (logits) of shape ``(n, d)``.
        T: A scarlar temperature to calibrate the score.
    
    Returns:
        MCP: The calculated MCP of shape ``(n, )``

    """
    p = cal_softmax(x, T = T)
    MCP = np.max(p, axis=1)

    return MCP

def cal_entropy(x: np.ndarray, T = 1, e1 = 1e-6) -> np.ndarray:
    """Compute entropy values for each sets of scores in x.

    Note:
        This entropy calculation chooses base as the number of class to ensure normalization.
    
    Args:
        x: The network output (logits) of shape ``(n, d)``.
        T: A scarlar temperature to calibrate the score.
        e1: A small number to prevent unexpected results of division.
    
    Returns:
        entropy: The calculated entropy of shape ``(n, )``

    """
    num_class = x.shape[-1]
    p = cal_softmax(x, T = T)
    entropy = - np.sum(p * np.emath.logn(num_class, p + e1), axis=1)

    return entropy

def cal_doctor(x: np.ndarray, T = 1) -> np.ndarray:
    """Compute engery values for each sets of scores in x.

    Calculating based on the defination as described in:
    ``https://arxiv.org/abs/2106.02395``
    
    Args:
        x: The network output (logits) of shape ``(n, d)``.
        T: A scarlar temperature to calibrate the score.
    
    Returns:
        doctor: The calculated doctor of shape ``(n, )``

    """

    p = cal_softmax(x, T = T)
    g = np.sum(p ** 2, axis=1)
    doctor = (1 - g) / g

    return doctor

def sum_tensor(inp: np.ndarray, axes: np.ndarray, keepdims=False) -> np.ndarray:
    axes = np.unique(axes).astype(int)
    if keepdims:
        for ax in axes:
            inp = inp.sum(int(ax), keepdims=True)
    else:
        for ax in sorted(axes, reverse=True):
            inp = inp.sum(int(ax))
    return inp

def get_tp_fp_fn(net_output: np.ndarray, gt: np.ndarray, axes=None, square=False) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Calculate the true positive (tp), false positive (fp) and false negative (fn).

    Args:
        net_output: The predicted probability of shape ``(n, d, (H), (W), (D))``.
        gt: The manual label of shape ``(n, (H), (W), (D))``.
        axes: We average the axes of features. This is useful if we want to return case-wise soft dice score.
        square: Indicate if we want to include the square term in tp etc, which could be more stable.
    
    Returns:
        tp: The calculated class-wise true positive of shape ``(n, d)`` or ``(d, )``.
        fp: The calculated class-wise false positive of shape ``(n, d)`` or ``(d, )``.
        fn: The calculated class-wise false negative of shape ``(n, d)`` or ``(d, )``.

    """
    if axes is None:
        axes = tuple(range(2, len(net_output.shape)))

    shp_x = net_output.shape
    shp_y = gt.shape

    y_onehot = one_hot_embedding(gt, shp_x[1])

    tp = net_output * y_onehot
    fp = net_output * (1 - y_onehot)
    fn = (1 - net_output) * y_onehot

    if square:
        tp = tp ** 2
        fp = fp ** 2
        fn = fn ** 2

    tp = sum_tensor(tp, axes, keepdims=False)
    fp = sum_tensor(fp, axes, keepdims=False)
    fn = sum_tensor(fn, axes, keepdims=False)

    return tp, fp, fn

def SoftDiceLoss(x: np.ndarray, y: np.ndarray, smooth = 1e-5) -> np.ndarray:
    """Calculation of soft dice score / F1 score.

    Args:
        x: The predicted probability of shape ``(n, d, (H), (W), (D))``.
        y: The manual label of shape ``(n, (H), (W), (D))``.
    
    Returns:
        dice_score: Average class-wise dice of shape ``(d, )``

    """
    assert len(x.shape) == len(y.shape) + 1

    shp_x = x.shape
    square = False

    axes = [0] + list(range(2, len(shp_x)))
    tp, fp, fn = get_tp_fp_fn(x, y, axes, square)
    dc = (2 * tp + smooth) / (2 * tp + fp + fn + smooth)
    dc_process = dc

    return dc_process

def SoftSensitivity(x: np.ndarray, y: np.ndarray, smooth = 1e-5) -> np.ndarray:
    """Calculation of soft sensitivity.

    Args:
        x: The predicted probability of shape ``(n, d, (H), (W), (D))``.
        y: The manual label of shape ``(n, (H), (W), (D))``.
    
    Returns:
        sensitivity_score: Average class-wise sensitivity of shape ``(d, )``

    """
    assert len(x.shape) == len(y.shape) + 1

    shp_x = x.shape
    square = False

    axes = [0] + list(range(2, len(shp_x)))
    tp, fp, fn = get_tp_fp_fn(x, y, axes, square)
    sensitivity = (tp + smooth) / (tp + fn + smooth)

    return sensitivity

def SoftPrecision(x: np.ndarray, y: np.ndarray, smooth = 1e-5) -> np.ndarray:
    """Calculation of soft precision score.

    Args:
        x: The predicted probability of shape ``(n, d, (H), (W), (D))``.
        y: The manual label of shape ``(n, (H), (W), (D))``.
    
    Returns:
        precision_score: Average class-wise precision of shape ``(d, )``

    """
    assert len(x.shape) == len(y.shape) + 1

    shp_x = x.shape
    square = False

    axes = [0] + list(range(2, len(shp_x)))
    tp, fp, fn = get_tp_fp_fn(x, y, axes, square)
    precision = (tp + smooth) / (tp + fp + smooth)

    return precision

def SoftAUC(x: np.ndarray, sel_cls: int = None) -> np.ndarray:
    """Calculate the estimated AUC given the probability.
    
    Note:
        When dealing with multi-class AUC, we calculate the One-vs-Rest strategy for simplicity.
        More information could refer to https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html.
        Similar strategy is adopted in https://nannyml.readthedocs.io/en/stable/how_it_works/performance_estimation.html.
    
    Args:
        x: The predicted probability of shape ``(n, d, (H), (W), (D))``.
        sel_cls: Selected class to calculate AUC.

    Returns:
        AUCs: Estimated class-wise AUC of shape ``(d, )``, or ``(1, )`` if sel_cls is givien.
        TPRall: A list of d true positive rates for different classes, or a list of 1 if sel_cls is givien.
        FPRall: A list of d false positive rates for different classes, or a list of 1 if sel_cls is givien.

    """
    
    # thresholds = [0.00001, 0.0001, 0.001, 0.01, 0.05, 
    #               0.1, 0.2, 0.3, 0.4, 0.5, 
    #               0.6, 0.7, 0.8, 0.9, 0.95, 
    #               0.99, 0.999, 0.9999, 0.99999]
    
    y = np.argmax(x, axis = 1) # this is not important, just a placeholder of shape ``(n, (H), (W), (D))``
    shp_x = x.shape
    square = False
    axes = [0] + list(range(2, len(shp_x)))
    #
    AUCs = []
    TPRall = []
    FPRall = []
    smooth = 1e-6
    #
    if sel_cls is None:
        allcls = range(x.shape[1])
    else:
        allcls = range(sel_cls, sel_cls+1)

    for test_cls in allcls:
        # low thres -> high FPR -> high sensitivity (TPR)
        # obtain TPRs and FPRs

        # customize thresholds
        # generate thresholds that fit the probability, just in case tpr == 0
        prob_max = np.min((np.max(x[:, test_cls]) - 0.05, 1.))
        prob_min = np.max((np.min(x[:, test_cls]) + 0.05, 0.))
        num_points = 5 # this might is too small, but I need to accelerate the optimization process.

        interval = (prob_max - prob_min) / (num_points - 1)
        thresholds = [prob_min + i * interval for i in range(num_points)]

        TPRs = []
        FPRs = []
        #
        TPRs.append(0.)
        FPRs.append(0.)
        for thres in thresholds:
            cls_choice = list(range(x.shape[1]))
            cls_choice.remove(test_cls)
            y[y == test_cls] = cls_choice[0] # remove all samples that predicted as ``test_cls``
            #
            mask = x[:, test_cls] > thres
            y[mask, ...] = test_cls # overwrite the prediction based on given threshold
            #
            # I only need to make sure y[y_cls == True] is correct, and ``test_cls`` does not appear otherwhere

            tp, fp, fn = get_tp_fp_fn(x, y, axes, square)

            TPR = tp[test_cls] / (np.sum(y == test_cls) + smooth)
            FPR = fp[test_cls] / (np.sum(y != test_cls) + smooth)

            TPRs.append(TPR)
            FPRs.append(FPR)
        TPRs.append(1.)
        FPRs.append(1.)
        # calculate AUC
        AUC = 0
        for k_thres in range(len(TPRs) - 1):
            TPR_left = TPRs[k_thres]
            TPR_right = TPRs[k_thres + 1]
            FPR_left = FPRs[k_thres]
            FPR_right = FPRs[k_thres + 1]
            AUC_k = (FPR_right - FPR_left) * (TPR_left + TPR_right) / 2
            AUC = AUC + AUC_k
        AUCs.append(AUC)
        TPRall.append(TPRs)
        FPRall.append(FPRs)
    
    return np.array(AUCs), TPRall, FPRall


def one_hot_embedding(labels: np.ndarray, num_class: int) -> np.ndarray:
    """Embedding labels to one-hot form.

    Args:
        labels: The label map of shape ``(n, )`` or ``(n, H, W, (D))``.
        num_class: An int scalar number of classes.

    Returns:
        labels_onehot: The transformed onehot label map of shape ``(n, d)`` or ``(n, d, H, W, (D))``.
    
    """
    y = np.eye(num_class)[labels.reshape(-1).astype(int)]  # of shape `(n * H * W (*D), d)`
    labels_onehot = y.reshape(list(labels.shape)+[num_class]) # of shape `(n, H, W, (D), d)`

    axes = [0] + [len(labels.shape)] + list(range(1, len(labels.shape)))
    labels_onehot = np.transpose(labels_onehot, axes) # of shape `(n, d, H, W, (D))`

    return labels_onehot


"""Define the MOVAL model."""
import numpy as np
from typing import Callable, Iterable, List, Literal, Optional, Tuple, Union
from sklearn.base import BaseEstimator
import itertools
import moval.models
import moval.solvers
import pickle

class MOVAL(BaseEstimator):
    """MOVAL model defined as part of a ``scikit-learn``-like API.

    Attributes:
        mode (str):
            The given task to estimate model performance. |Default:| ``classification``
        metric (str):
            The performance metric we optimize moval model to align. |Default:| ``accuracy``
            Metric can be ``accuracy`` | ``sensitivity`` | ``precision`` | ``f1score`` | ``auc``
            This is only effective when class_specific is ``True``, otherwise moval will always align overall accuracy.
        confidence_scores (str):
            The method to calculate the confidence scores. We provide a list of confidence score calculation methods which 
            can be displayed by running :py:func:`moval.models.get_conf_options`. |Default:| ``max_class_probability-conf``
        estim_algorithm (str):
            The algorithm to estimate model performance. We also provide a list of estimation algorithm which can be displayed by
            running :py:func:`moval.models.get_estim_options`. |Default:| ``ac-model``
        class_specific (bool):
            If ``True``, the calculation will match class-wise confidence to class-wise accuracy. |Default:| ``False``
        approximate (bool):
            If ``True``, we crop the image and label map to accelerate the optimization of segmentation. |Default:| ``False``
        approximate_boundary (int):
            The enlarged regions of the region to crop. |Default:| ``30``

    Example:
        
        >>> import moval
        >>> moval_model = moval.MOVAL(
                            mode = "classification",
                            metric = "accuracy",
                            confidence_scores = "max_class_probability-conf",
                            estim_algorithm = "ac-model",
                            class_specific = False
                            )

    """


    def __init__(
        self,
        mode: str = "classification",
        metric: str = "accuracy",
        confidence_scores: str = "max_class_probability-conf",
        estim_algorithm: str = "ac-model",
        class_specific: bool = False, 
        approximate: bool = False, 
        approximate_boundary: int = 30
    ):
        self.__dict__.update(locals())

        # decide if the model is in the ensemble mode
        if self.estim_algorithm.split('-')[1] == 'ensemble':
            self.ensemble = True
        else:
            self.ensemble = False

    def fit(
        self,
        logits: Union[List[Iterable], np.ndarray],
        gt: Union[List[Iterable], np.ndarray], 
        batch: int = 1
    ) -> "MOVAL":
        """Fit the estimator to the given dataset by minimzing the calibration error.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` for classification and a list of n ``(d, H, W, (D))`` for segmentation. 
            gt: The cooresponding annotation of shape ``(n, )`` for classification and a list of n ``(H, W, (D))`` for segmentation.
            batch: To match the group-wise accuracy with group-wise confidence score. batch is the group size. This is useful when validation data per case is few.
        
        Note:
            It only makes sense if the class frequency is similar between class index to use ``batch`` > 1!
        
        Return:
            ``self``

        Example:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)

        """

        # Input validation
        if self.mode == "classification":
            if logits.shape[0] != gt.shape[0]:
                raise ValueError(
                        f"Invalid input samples: logits and GT should represent the same number of samples"
                        f"(n_samples of logits, {logits.shape[0]}, while n_samples of GT , {gt.shape[0]})."
                    )

            if len(logits.shape) != (len(gt.shape) + 1):
                raise ValueError(
                        f"Invalid input dimension: logits should have one additional dimension than GT"
                        f"(dimension of logits, {len(logits.shape)}, while dimension of GT , {len(gt.shape)})."
                    )
            
            if len(gt.shape) != 1:
                raise ValueError(
                    f"Invalid input dimension for classification: GT should have the dimension of 1"
                    f"(dimension of GT, {len(logits.shape)})."
                )

        else:
            if len(logits) != len(gt):
                raise ValueError(
                        f"Invalid input samples: logits and GT should represent the same number of samples"
                        f"(n_samples of logits, {len(logits)}, while n_samples of GT , {len(gt)})."
                    )

            if len(logits[0].shape) != (len(gt[0].shape) + 1):
                raise ValueError(
                        f"Invalid input dimension: logits should have one additional dimension than GT"
                        f"(dimension of logits, {len(logits[0].shape)}, while dimension of GT , {len(gt[0].shape)})."
                    )
            
            if (len(gt[0].shape) != 2) and (len(gt[0].shape) != 3):
                raise ValueError(
                    f"Invalid input dimension for segmentation: GT should have the dimension of 2 or 3"
                    f"(dimension of GT, {len(gt[0].shape)})."
                )

        if self.mode == "segmentation" and self.approximate == True:
            logits, gt = self.crop(logits = logits, gt = gt, approximate_boundary = self.approximate_boundary)
        
        # initilization
        if isinstance(logits, list):
            self.numclass = logits[0].shape[0]
        else:
            self.numclass = logits.shape[-1]

        if self.estim_algorithm == "moval-ensemble":
            # All the CS variants
            moval_models = list(itertools.product(
                ["ts-model", "doc-model", "atc-model", "ts-atc-model"],
                [self.mode],
                moval.models.get_conf_options(),
                [True]))
        elif self.estim_algorithm == "moval-ensemble-triathlon":
            # All the CS variants that support the estimation of all metrics
            # cannot utilize atc model family as they cannot estimate false positives.
            moval_models = list(itertools.product(
                ["ts-model", "doc-model"],
                [self.mode],
                moval.models.get_conf_options(),
                [True]))
        elif self.estim_algorithm == "moval-ensemble-debug":
            # just for debuging, aggregating 2 conditions.
            moval_models = []
            moval_models.append(["ts-model", self.mode, "energy-conf", True])
            moval_models.append(["doc-model", self.mode, "max_class_probability-conf", False])
        # classification policy here
        elif self.estim_algorithm == "moval-ensemble-cls-accuracy":
            # ensenble strategies for accuracy estimation of classification
            moval_models = []
            moval_models.append(["ts-atc-model", self.mode, "max_class_probability-conf", True])
            moval_models.append(["ts-atc-model", self.mode, "entropy-conf", True])
            moval_models.append(["atc-model", self.mode, "energy-conf", True])
        elif self.estim_algorithm == "moval-ensemble-cls-f1score":
            # ensenble strategies for f1score estimation of classification
            moval_models = []
            moval_models.append(["atc-model", self.mode, "doctor-conf", True])
            moval_models.append(["atc-model", self.mode, "entropy-conf", True])
            moval_models.append(["atc-model", self.mode, "max_class_probability-conf", True])
        elif self.estim_algorithm == "moval-ensemble-cls-sensitivity":
            # ensenble strategies for sensitivity estimation of classification
            moval_models = []
            moval_models.append(["ts-atc-model", self.mode, "doctor-conf", True])
            moval_models.append(["atc-model", self.mode, "entropy-conf", True])
            moval_models.append(["atc-model", self.mode, "max_class_probability-conf", True])
        elif self.estim_algorithm == "moval-ensemble-cls-precision":
            # ensenble strategies for precision estimation of classification
            moval_models = []
            moval_models.append(["ts-model", self.mode, "entropy-conf", True])
            moval_models.append(["doc-model", self.mode, "max_class_probability-conf", True])
            moval_models.append(["doc-model", self.mode, "entropy-conf", True])
        elif self.estim_algorithm == "moval-ensemble-cls-auc":
            # ensenble strategies for auc estimation of classification
            moval_models = []
            moval_models.append(["ts-model", self.mode, "energy-conf", False])
            moval_models.append(["ts-model", self.mode, "max_class_probability-conf", True])
            moval_models.append(["doc-model", self.mode, "energy-conf", True])
        # segmentation 2d policy here
        elif self.estim_algorithm == "moval-ensemble-seg2d-accuracy":
            # ensenble strategies for accuracy estimation of segmentation
            moval_models = []
            moval_models.append(["atc-model", self.mode, "max_class_probability-conf", False])
            moval_models.append(["doc-model", self.mode, "entropy-conf", True])
            moval_models.append(["ts-model", self.mode, "max_class_probability-conf", True])
        elif self.estim_algorithm == "moval-ensemble-seg2d-f1score":
            # ensenble strategies for f1score estimation of segmentation
            moval_models = []
            moval_models.append(["doc-model", self.mode, "entropy-conf", True])
            moval_models.append(["doc-model", self.mode, "doctor-conf", True])
            moval_models.append(["atc-model", self.mode, "max_class_probability-conf", True])
        elif self.estim_algorithm == "moval-ensemble-seg2d-sensitivity":
            # ensenble strategies for sensitivity estimation of segmentation
            moval_models = []
            moval_models.append(["atc-model", self.mode, "max_class_probability-conf", False])
            moval_models.append(["ts-atc-model", self.mode, "max_class_probability-conf", False])
            moval_models.append(["atc-model", self.mode, "doctor-conf", False])
        elif self.estim_algorithm == "moval-ensemble-seg2d-precision":
            # ensenble strategies for precision estimation of segmentation
            moval_models = []
            moval_models.append(["ts-model", self.mode, "doctor-conf", True])
            moval_models.append(["doc-model", self.mode, "max_class_probability-conf", True])
            moval_models.append(["doc-model", self.mode, "doctor-conf", True])
        elif self.estim_algorithm == "moval-ensemble-seg2d-auc":
            # ensenble strategies for auc estimation of segmentation
            moval_models = []
            moval_models.append(["ts-model", self.mode, "entropy-conf", True])
            moval_models.append(["ts-model", self.mode, "doctor-conf", True])
            moval_models.append(["doc-model", self.mode, "doctor-conf", True])
        # segmentation 3d policy here
        elif self.estim_algorithm == "moval-ensemble-seg3d-accuracy":
            # ensenble strategies for accuracy estimation of segmentation
            moval_models = []
            moval_models.append(["ts-model", self.mode, "max_class_probability-conf", False])
            moval_models.append(["ts-model", self.mode, "max_class_probability-conf", True])
            moval_models.append(["doc-model", self.mode, "doctor-conf", True])
        elif self.estim_algorithm == "moval-ensemble-seg3d-f1score":
            # ensenble strategies for f1score estimation of segmentation
            moval_models = []
            moval_models.append(["ts-model", self.mode, "doctor-conf", True])
            moval_models.append(["ts-model", self.mode, "entropy-conf", True])
            moval_models.append(["ts-atc-model", self.mode, "energy-conf", True])
        elif self.estim_algorithm == "moval-ensemble-seg3d-sensitivity":
            # ensenble strategies for sensitivity estimation of segmentation
            moval_models = []
            moval_models.append(["doc-model", self.mode, "energy-conf", True])
            moval_models.append(["ts-model", self.mode, "max_class_probability-conf", True])
            moval_models.append(["ts-atc-model", self.mode, "energy-conf", True])
        elif self.estim_algorithm == "moval-ensemble-seg3d-precision":
            # ensenble strategies for precision estimation of segmentation
            moval_models = []
            moval_models.append(["ts-model", self.mode, "max_class_probability-conf", True])
            moval_models.append(["doc-model", self.mode, "max_class_probability-conf", True])
            moval_models.append(["ts-model", self.mode, "doctor-conf", False])
        elif self.estim_algorithm == "moval-ensemble-seg3d-auc":
            # ensenble strategies for auc estimation of segmentation
            moval_models = []
            moval_models.append(["ts-model", self.mode, "max_class_probability-conf", True])
            moval_models.append(["ts-model", self.mode, "max_class_probability-conf", False])
            moval_models.append(["doc-model", self.mode, "max_class_probability-conf", True])


        if self.ensemble:
            ensemble_conds = []
            models = []
            solvers = []
            probabilities = []
            for estim_algorithm, mode, confidence_scores, class_specific in moval_models:
                model = moval.models.init(
                    estim_algorithm,
                    mode = mode,
                    num_class = self.numclass,
                    confidence_scores = confidence_scores,
                    class_specific = class_specific
                )

                if class_specific == True and self.metric == "auc":
                    # find a good initatlization first for auc, as it could unstable
                    model_pre = moval.models.init(
                        estim_algorithm,
                        mode = self.mode,
                        num_class = self.numclass,
                        confidence_scores = self.confidence_scores,
                        class_specific = True
                        )
                    solver_pre = moval.solvers.init("base-solver", model = model_pre, metric = "precision")
                    solver_pre.fit(logits, gt)
                    model.param = model_pre.param

                solver = moval.solvers.init("base-solver", model = model, metric = self.metric)
                solver.fit(logits, gt, batch) # model fitting
                if self.metric == "precision" or self.metric == "auc":
                    probability_cond = model.calculate_probability(logits, appr = True, full = True) # ``(n, d)`` or a list of n ``(d, H, W, (D))``
                else:
                    probability_cond = model.calculate_probability(logits, appr = True) # ``(n, d)`` or a list of n ``(d, H, W, (D))``
                #
                models.append(model)
                solvers.append(solver)
                probabilities.append(probability_cond)
                ensemble_cond = f"estim_algorithm_{estim_algorithm}_mode_{mode}_confidence_scores_{confidence_scores}_class_specific_{class_specific}"
                ensemble_conds.append(ensemble_cond)
            #
            # aggregate probabilities
            probability = self.probability_aggregation(probabilities)
            self.model_ = models
            self.solver_ = solvers
            print(f"Calculating and saving the fitted case-wise performance...")
            self.fitted_perf = self.get_case_perf(models[0], self.metric, logits, probability, gt)
            self.ensemble_conds = ensemble_conds

        else:
            model = moval.models.init(
                self.estim_algorithm,
                mode = self.mode,
                num_class = self.numclass,
                confidence_scores = self.confidence_scores,
                class_specific = self.class_specific
                )
            
            if self.class_specific == True and self.metric == "auc":
                # find a good initatlization first for auc, as it could unstable
                model_pre = moval.models.init(
                    self.estim_algorithm,
                    mode = self.mode,
                    num_class = self.numclass,
                    confidence_scores = self.confidence_scores,
                    class_specific = True
                    )
                solver_pre = moval.solvers.init("base-solver", model = model_pre, metric = "precision")
                solver_pre.fit(logits, gt)
                model.param = model_pre.param

            solver = moval.solvers.init("base-solver", model = model, metric = self.metric)
            solver.fit(logits, gt, batch) # model fitting

            # save the results to self attributes.
            self.model_ = model
            self.solver_ = solver
            print(f"Calculating and saving the fitted case-wise performance...")
            if self.metric == "precision" or self.metric == "auc":
                probability = model.calculate_probability(logits, appr = True, full=True)
            else:
                probability = model.calculate_probability(logits, appr = True)
            self.fitted_perf = self.get_case_perf(model, self.metric, logits, probability, gt)

        if self.mode == "classification":
            self.n_dim_ = len(logits.shape)
        else:
            self.n_dim_ = len(logits[0].shape)

        return self

    @classmethod
    def probability_aggregation(cls,
                                probabilities):
        """Calcualte the mean of multiple probabilities.

        Args:
            probabilities: The predicted probability.
                This is a list of k ``(n, d)`` for clasfication and a list of k lists of n ``(d, H, W, (D))`` for segmentation.
        
        Returns:
            probability_agg: The average probability of shape ``(n, d)`` for classification and a list of n ``(d, H, W, (D))`` for segmentation. 

        """
        if isinstance(probabilities[0], list):
            # segmentation
            probability_agg = []
            for n_case in range(len(probabilities[0])):
                probabilities_case = []
                for k_cond in range(len(probabilities)):
                    probabilities_case.append(probabilities[k_cond][n_case])
                probability_agg.append(np.mean(np.array(probabilities_case), axis = 0))
        else:
            # classification
            probability_agg = np.mean(np.array(probabilities), axis = 0)
        
        return probability_agg

    @classmethod
    def get_case_perf(cls,
                      model: moval.models,
                      metric: str,
                      inp: Union[List[Iterable], np.ndarray],
                      probability: Union[List[Iterable], np.ndarray],
                      gt: Union[List[Iterable], np.ndarray]):
        """Store the estimated results of n fitted data.

        Note:
            For segmentaiton tasks, we only save the average dsc of all the foreground classes.

        Args:
            model: The fitted moval model.
            metric: The performance metric to follow.
            inp: The network output (logits) of shape ``(n, d)`` for classification and a list of n ``(d, H, W, (D))`` for segmentation.
            probability: The calculated probability of shape ``(n, d)`` for classification and a list of n ``(d, H, W, (D))`` for segmentation.
            gt: The cooresponding annotation of shape ``(n, )`` for classification and a list of n ``(H, W, (D))`` for segmentation.

        Returns:
            fitted_perf: a list contain estimated results of shape of len n. Each element is a scalar if estimated accuracy, otherwise ``(d, )``.
        
        """

        fitted_perf = []

        if model.mode == "classification":
            
            for n_case in range(len(inp)):

                if metric == "accuracy":
                    estim_perf_case, _ = model.estimate_accuracy(inp[n_case:n_case + 1])
                elif metric == "sensitivity":
                    estim_perf_case = model.estimate_sensitivity(inp[n_case:n_case + 1], probability[n_case:n_case + 1])
                elif metric == "precision":
                    estim_perf_case = model.estimate_precision(probability[n_case:n_case + 1])
                elif metric == "f1score":
                    estim_perf_case = model.estimate_f1score(inp[n_case:n_case + 1], probability[n_case:n_case + 1])
                elif metric == "auc":
                    estim_perf_case = model.estimate_auc(probability[n_case:n_case + 1])
                else:
                    ValueError(f"Unsupported metric '{metric}'") 
                
                fitted_perf.append(estim_perf_case)
        
        else:

            # generate gt_guide from gt here.
            gt_guide = []
            for n_case in range(len(gt)):
                gt_case     = gt[n_case]
                gt_exist = []
                for k_cls in range(model.num_class):
                    gt_exist.append(np.sum(gt_case == k_cls) > 0)
                gt_guide.append(gt_exist)
            gt_guide = np.array(gt_guide)

            for n_case in range(len(inp)):
                
                if metric == "accuracy":
                    estim_perf_case, _ = model.estimate_accuracy(inp[n_case:n_case + 1], gt_guide = gt_guide[n_case:n_case + 1])
                elif metric == "sensitivity":
                    estim_perf_case = model.estimate_sensitivity(inp[n_case:n_case + 1], probability[n_case:n_case + 1], gt_guide = gt_guide[n_case:n_case + 1])
                elif metric == "precision":
                    estim_perf_case = model.estimate_precision(probability[n_case:n_case + 1], gt_guide = gt_guide[n_case:n_case + 1])
                elif metric == "f1score":
                    estim_perf_case = model.estimate_f1score(inp[n_case:n_case + 1], probability[n_case:n_case + 1], gt_guide = gt_guide[n_case:n_case + 1])
                elif metric == "auc":
                    estim_perf_case = model.estimate_auc(probability[n_case:n_case + 1], gt_guide = gt_guide[n_case:n_case + 1])
                else:
                    ValueError(f"Unsupported metric '{metric}'")
                
                fitted_perf.append(estim_perf_case)
        
        return fitted_perf

    @classmethod
    def crop(cls,
             logits: List[Iterable],
             gt: List[Iterable],
             approximate_boundary):
        """Crop the image and label map to accelrate the optimization process.
        
        Here we do the cropping based on the label map (gt). We first find the bounding box and enlarge the region with boundary = 30 pixels.

        Args:
            logits: The network output (logits) of a list of n ``(d, H, W, (D))``. 
            gt: The cooresponding annotation of a list of n ``(H, W, (D))``.
            approximate_boundary: The enlarged regions of the region to crop.
        
        Return:
            logits_post: The cropped network output (logits) of a list of n ``(d, H', W', (D'))``. 
            gt_post: The cooresponding cropped annotation of a list of n ``(H', W', (D'))``.

        """

        logits_post = []
        gt_post = []
        for k_case in range(len(logits)):
            logit_map = logits[k_case]
            gt_map = gt[k_case]
        
            if len(gt_map.shape) == 2:
                border_x, border_y = gt_map.shape
                ind_x, ind_y = np.where(gt_map != 0)
                #
                max_x = np.min((np.max(ind_x) + approximate_boundary, border_x))
                min_x = np.max((np.min(ind_x) - approximate_boundary, 0))
                #
                max_y = np.min((np.max(ind_y) + approximate_boundary, border_y))
                min_y = np.max((np.min(ind_y) - approximate_boundary, 0))
                #
                logits_post.append(logit_map[:, min_x:max_x, min_y:max_y])
                gt_post.append(gt_map[min_x:max_x, min_y:max_y])

            elif len(gt_map.shape) == 3:
                border_x, border_y, border_z = gt_map.shape
                ind_x, ind_y, ind_z = np.where(gt_map != 0)
                #
                max_x = np.min((np.max(ind_x) + approximate_boundary, border_x))
                min_x = np.max((np.min(ind_x) - approximate_boundary, 0))
                #
                max_y = np.min((np.max(ind_y) + approximate_boundary, border_y))
                min_y = np.max((np.min(ind_y) - approximate_boundary, 0))
                #
                max_z = np.min((np.max(ind_z) + approximate_boundary, border_z))
                min_z = np.max((np.min(ind_z) - approximate_boundary, 0))
                #
                logits_post.append(logit_map[:, min_x:max_x, min_y:max_y, min_z:max_z])
                gt_post.append(gt_map[min_x:max_x, min_y:max_y, min_z:max_z])
            else:
                raise ValueError("Not implemented!")
        
        return logits_post, gt_post

    def estimate(self,
                 logits: Union[List[Iterable], np.ndarray],
                 gt_guide: np.ndarray = None):
        """The function to estimate the model performance. This will call other corresponding metric calculation functions.
        
        Args:
            logits: A numpy array of size ``(n, d)`` for classification or a list of n ``(d, H, W, (D))`` for segmentation.
            gt_guide: A numpy array of size ``(n, d)`` for segmentation, indicating the existing of object d in sample n.
                If ``False``, it means that there isn't any manuel label of class d in this sample.
        
        Returns:
            estim: estimated metric (float) for classification tasks, 
                or class-wise estimated metric of shape ``(d-1, )`` for segmentation tasks.

        Example:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)
            >>> estim_acc = moval_model.estimate(logits)

        """

        if self.metric == "accuracy":
            return self.estimate_accuracy(logits, gt_guide)
        elif self.metric == "sensitivity":
            return self.estimate_sensitivity(logits, gt_guide)
        elif self.metric == "precision":
            return self.estimate_precision(logits, gt_guide)
        elif self.metric == "f1score":
            return self.estimate_f1score(logits, gt_guide)
        elif self.metric == "auc":
            return self.estimate_auc(logits, gt_guide)
        else:
            ValueError(f"Unsupported metric '{self.metric}'")

    def get_probability(self,
                        logits: Union[List[Iterable], np.ndarray]):
        """Obtain the the probability give the logits.

        Note:
            It will ensemble (not exactly ensemble, but calculate the mean of probability) moval-ensemble.
            Here we utilize ``calculate_probability`` for classification, while ``calculate_probability_appr`` for segmentation. 
            Because the sample number of segmentation tasks is too large.

        Args:
            logits: A numpy array of size ``(n, d)`` for classification or a list of n ``(d, H, W, (D))`` for segmentation.
        
        Examples:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> moval_model = moval.MOVAL(confidence_scores = "moval-ensemble")
            >>> probability = moval_model.get_probability(logits)
        
        """
            
        if self.ensemble:
            probabilities = []
            for model in self.model_:
                if self.metric == "precision" or self.metric == "auc":
                    probability_cond = model.calculate_probability(logits, appr = True, full=True)
                else:
                    probability_cond = model.calculate_probability(logits, appr = True)

                probabilities.append(probability_cond)
            probability = self.probability_aggregation(probabilities)
        else:
            model = self.model_
            if self.metric == "precision" or self.metric == "auc":
                probability = model.calculate_probability(logits, appr = True, full=True)
            else:
                probability = model.calculate_probability(logits, appr = True)

        return probability

    def estimate_accuracy(self,
                 logits: Union[List[Iterable], np.ndarray],
                 gt_guide: np.ndarray = None):
        """Estimate accuracy using logits.

        Args:
            logits: A numpy array of size ``(n, d)`` for classification or a list of n ``(d, H, W, (D))`` for segmentation.
            gt_guide: A numpy array of size ``(n, d)`` for segmentation, indicating the existing of object d in sample n.
                If ``False``, it means that there isn't any manuel label of class d in this sample.
        
        Returns:
            estim: estimated accuracy (float). 

        Example:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)
            >>> estim_acc = moval_model.estimate_accuracy(logits)

        """

        # Input validation
        if self.mode == "classification":
            if self.n_dim_ != len(logits.shape):
                raise ValueError(
                        f"Inconsistent input dimension: training and test samples have different dimenstions"
                        f"(dimension of training samples, {len(self.n_dim_)}, while dimension of test samples, {len(logits.shape)})"
                    )
        else:
            if self.n_dim_ != len(logits[0].shape):
                raise ValueError(
                        f"Inconsistent input dimension: training and test samples have different dimenstions"
                        f"(dimension of training samples, {len(self.n_dim_)}, while dimension of test samples, {len(logits[0].shape)})"
                    )

        if self.ensemble:
            
            probability = self.get_probability(logits)

            if self.mode == "classification":
                estim_acc = np.mean(np.max(probability, axis = 1))
                return estim_acc
            else:
                scores = []
                for n_case in range(len(probability)):
                    score_flatten = np.max(probability[n_case], axis = 0).flatten() # ``n``
                    scores.append(score_flatten)
                estim_acc = np.mean( np.concatenate(scores) )
                
                return estim_acc
        else:
            model = self.model_

            estim_acc, _ = model.estimate_accuracy(logits)
            return estim_acc
    
    def estimate_sensitivity(self,
                 logits: Union[List[Iterable], np.ndarray],
                 gt_guide: np.ndarray = None):
        """Estimate model's sensitivity using logits.

        Args:
            logits: A numpy array of size ``(n, d)`` for classification or a list of n ``(d, H, W, (D))`` for segmentation.
            gt_guide: A numpy array of size ``(n, d)`` for segmentation, indicating the existing of object d in sample n.
                If ``False``, it means that there isn't any manuel label of class d in this sample.
        
        Returns:
            estim: estimated sensitivity of shape ``(d, )``.
        
        Example:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)
            >>> estim_sensitivity = moval_model.estimate_sensitivity(logits)

        """

        # Input validation
        if self.mode == "classification":
            if self.n_dim_ != len(logits.shape):
                raise ValueError(
                        f"Inconsistent input dimension: training and test samples have different dimenstions"
                        f"(dimension of training samples, {len(self.n_dim_)}, while dimension of test samples, {len(logits.shape)})"
                    )
        else:
            if self.n_dim_ != len(logits[0].shape):
                raise ValueError(
                        f"Inconsistent input dimension: training and test samples have different dimenstions"
                        f"(dimension of training samples, {len(self.n_dim_)}, while dimension of test samples, {len(logits[0].shape)})"
                    )

        probability = self.get_probability(logits)
        if self.ensemble:
            estim_sensitivity = self.model_[0].estimate_sensitivity(logits, probability, gt_guide = gt_guide)
        else:
            estim_sensitivity = self.model_.estimate_sensitivity(logits, probability, gt_guide = gt_guide)

        return estim_sensitivity
    
    def estimate_precision(self,
                 logits: Union[List[Iterable], np.ndarray],
                 gt_guide: np.ndarray = None):
        """Estimate model's sensitivity using logits.

        Args:
            logits: A numpy array of size ``(n, d)`` for classification or a list of n ``(d, H, W, (D))`` for segmentation.
            gt_guide: A numpy array of size ``(n, d)`` for segmentation, indicating the existing of object d in sample n.
                If ``False``, it means that there isn't any manuel label of class d in this sample.
        
        Returns:
            estim: estimated average precision (float) for classification tasks, 
                or estimated precision of shape ``(d-1, )`` for segmentation tasks.

        Example:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)
            >>> estim_precision = moval_model.estimate_precision(logits)

        """

        # Input validation
        if self.mode == "classification":
            if self.n_dim_ != len(logits.shape):
                raise ValueError(
                        f"Inconsistent input dimension: training and test samples have different dimenstions"
                        f"(dimension of training samples, {len(self.n_dim_)}, while dimension of test samples, {len(logits.shape)})"
                    )
        else:
            if self.n_dim_ != len(logits[0].shape):
                raise ValueError(
                        f"Inconsistent input dimension: training and test samples have different dimenstions"
                        f"(dimension of training samples, {len(self.n_dim_)}, while dimension of test samples, {len(logits[0].shape)})"
                    )

        probability = self.get_probability(logits)
        if self.ensemble:
            estim_precision = self.model_[0].estimate_precision(probability, gt_guide = gt_guide)
        else:
            estim_precision = self.model_.estimate_precision(probability, gt_guide = gt_guide)

        return estim_precision
    
    def estimate_f1score(self,
                 logits: Union[List[Iterable], np.ndarray],
                 gt_guide: np.ndarray = None):
        """Estimate model's F1score using logits.

        Args:
            logits: A numpy array of size ``(n, d)`` for classification or a list of n ``(d, H, W, (D))`` for segmentation.
            gt_guide: A numpy array of size ``(n, d)`` for segmentation, indicating the existing of object d in sample n.
                If ``False``, it means that there isn't any manuel label of class d in this sample.
        
        Returns:
            estim: estimated average F1score (float) for classification tasks, 
                or estimated dsc of shape ``(d-1, )`` for segmentation tasks.

        Example:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)
            >>> estim_f1score = moval_model.estimate_f1score(logits)
                
        """

        # Input validation
        if self.mode == "classification":
            if self.n_dim_ != len(logits.shape):
                raise ValueError(
                        f"Inconsistent input dimension: training and test samples have different dimenstions"
                        f"(dimension of training samples, {len(self.n_dim_)}, while dimension of test samples, {len(logits.shape)})"
                    )
        else:
            if self.n_dim_ != len(logits[0].shape):
                raise ValueError(
                        f"Inconsistent input dimension: training and test samples have different dimenstions"
                        f"(dimension of training samples, {len(self.n_dim_)}, while dimension of test samples, {len(logits[0].shape)})"
                    )

        probability = self.get_probability(logits)
        if self.ensemble:
            estim_f1score = self.model_[0].estimate_f1score(logits, probability, gt_guide = gt_guide)
        else:
            estim_f1score = self.model_.estimate_f1score(logits, probability, gt_guide = gt_guide)

        return estim_f1score

    def estimate_auc(self,
                 logits: Union[List[Iterable], np.ndarray],
                 gt_guide: np.ndarray = None):
        """Estimate model's AUC using logits.

        Args:
            logits: A numpy array of size ``(n, d)`` for classification or a list of n ``(d, H, W, (D))`` for segmentation.
            gt_guide: A numpy array of size ``(n, d)`` for segmentation, indicating the existing of object d in sample n.
                If ``False``, it means that there isn't any manuel label of class d in this sample.
        
        Returns:
            estim: estimated average AUC (float) for classification tasks, 
                or estimated AUC of shape ``(d-1, )`` for segmentation tasks.

        Example:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)
            >>> estim_auc = moval_model.estimate_auc(logits)
                
        """

        # Input validation
        if self.mode == "classification":
            if self.n_dim_ != len(logits.shape):
                raise ValueError(
                        f"Inconsistent input dimension: training and test samples have different dimenstions"
                        f"(dimension of training samples, {len(self.n_dim_)}, while dimension of test samples, {len(logits.shape)})"
                    )
        else:
            if self.n_dim_ != len(logits[0].shape):
                raise ValueError(
                        f"Inconsistent input dimension: training and test samples have different dimenstions"
                        f"(dimension of training samples, {len(self.n_dim_)}, while dimension of test samples, {len(logits[0].shape)})"
                    )

        probability = self.get_probability(logits)
        if self.ensemble:
            estim_auc = self.model_[0].estimate_auc(probability, gt_guide = gt_guide)
        else:
            estim_auc = self.model_.estimate_auc(probability, gt_guide = gt_guide)

        return estim_auc
    
    def save(self, filename: str):
        """Save current parameters to disk using pickle.

        Args:
            filename: The path to the file in which to save the parameters.
        
        Return:
            The saved parameters, or a list of saved paramters when requiring model ensembling.

        Examples:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)
            >>> moval_model.save('./foo.pkl')

        """

        if self.ensemble:
            # Object or variable to be saved
            name_to_save = filename[:-4]
            ckpts = []
            # A ckpt saving the meta-info
            ckpt = {}
            ckpt["mode"] = self.mode
            ckpt["metric"] = self.metric
            ckpt["estim_algorithm"] = self.estim_algorithm
            ckpt["mode"] = self.mode
            ckpt["confidence_scores"] = self.confidence_scores
            ckpt["class_specific"] = self.class_specific
            ckpt["ensemble"] = self.ensemble
            ckpt["ensemble_conds"] = self.ensemble_conds
            # Save the object as a pickle file
            with open(filename, 'wb') as file:
                pickle.dump(ckpt, file)
            ckpts.append(ckpt)
            for model, ensemble_cond in zip(self.model_, self.ensemble_conds):

                ckpt = {}

                ckpt["estim_algorithm"] = model.estim_algorithm
                ckpt["mode"] = model.mode
                ckpt["metric"] = self.metric
                ckpt["num_class"] = model.num_class
                ckpt["confidence_scores"] = model.confidence_scores
                ckpt["class_specific"] = model.class_specific
                ckpt["param"] = model.param
                ckpt["n_dim_"] = self.n_dim_
                ckpt["fitted_perf"] = self.fitted_perf
                if model.extend_param:
                    ckpt["param_ext"] = model.param_ext
                if model.conf.normalization:
                    ckpt["max_value"] = model.max_value
                    ckpt["min_value"] = model.min_value

                filename_to_save = f"{name_to_save}_{ensemble_cond}.pkl"

                # Save the object as a pickle file
                with open(filename_to_save, 'wb') as file:
                    pickle.dump(ckpt, file)

                ckpts.append(ckpt)

            return ckpts
        
        else:
            # Object or variable to be saved
            ckpt = {}
            
            ckpt["estim_algorithm"] = self.model_.estim_algorithm
            ckpt["mode"] = self.model_.mode
            ckpt["metric"] = self.metric
            ckpt["num_class"] = self.model_.num_class
            ckpt["confidence_scores"] = self.model_.confidence_scores
            ckpt["class_specific"] = self.model_.class_specific
            ckpt["param"] = self.model_.param
            ckpt["n_dim_"] = self.n_dim_
            ckpt["fitted_perf"] = self.fitted_perf
            ckpt["ensemble"] = self.ensemble
            if self.model_.extend_param:
                ckpt["param_ext"] = self.model_.param_ext
            if self.model_.conf.normalization:
                ckpt["max_value"] = self.model_.max_value
                ckpt["min_value"] = self.model_.min_value

            # Save the object as a pickle file
            with open(filename, 'wb') as file:
                pickle.dump(ckpt, file)

            return ckpt
    
    @classmethod
    def load(cls, filename: str) -> "MOVAL":
        """load parameters from disk.

        Args:
            filename: The path to the file in which to save the parameters.
        
        Return:
            The loaded model, or a list of loaded models when requiring model ensembling.
        
        Example:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)
            >>> moval_model.save('./foo.pkl')
            >>> loaded_model = moval.MOVAL.load('./foo.pkl')
            >>> estim_acc = loaded_model.estimate(logits)

        """

        with open(filename, 'rb') as file:
            loaded_ckpt = pickle.load(file)
        
        moval_model = moval.MOVAL(
                            mode = loaded_ckpt["mode"],
                            confidence_scores = loaded_ckpt["confidence_scores"],
                            estim_algorithm = loaded_ckpt["estim_algorithm"],
                            class_specific = loaded_ckpt["class_specific"]
                            )
        
        moval_model.ensemble = loaded_ckpt["ensemble"]

        if moval_model.ensemble:

            name_to_load = filename[:-4]
            moval_model.ensemble_conds = loaded_ckpt["ensemble_conds"]
            models = []
            for ensemble_cond in moval_model.ensemble_conds:
                filename_to_load = f"{name_to_load}_{ensemble_cond}.pkl"
                with open(filename_to_load, 'rb') as file:
                    loaded_ckpt = pickle.load(file)

                model = moval.models.init(
                    loaded_ckpt["estim_algorithm"],
                    mode = loaded_ckpt["mode"],
                    num_class = loaded_ckpt["num_class"],
                    confidence_scores = loaded_ckpt["confidence_scores"],
                    class_specific = loaded_ckpt["class_specific"]
                    )

                model.param = loaded_ckpt["param"]
                if model.extend_param:
                    model.param_ext = loaded_ckpt["param_ext"]
                
                if model.conf.normalization:
                    model.max_value = loaded_ckpt["max_value"]
                    model.min_value = loaded_ckpt["min_value"]

                model.is_fitted = True
                models.append(model)

                moval_model.model_ = models

        else:

            model = moval.models.init(
                loaded_ckpt["estim_algorithm"],
                mode = loaded_ckpt["mode"],
                num_class = loaded_ckpt["num_class"],
                confidence_scores = loaded_ckpt["confidence_scores"],
                class_specific = loaded_ckpt["class_specific"]
                )

            model.param = loaded_ckpt["param"]
            if model.extend_param:
                model.param_ext = loaded_ckpt["param_ext"]
            
            if model.conf.normalization:
                model.max_value = loaded_ckpt["max_value"]
                model.min_value = loaded_ckpt["min_value"]

            model.is_fitted = True

            moval_model.model_ = model

        moval_model.n_dim_ = loaded_ckpt["n_dim_"]
        moval_model.fitted_perf = loaded_ckpt["fitted_perf"]
        moval_model.metric = loaded_ckpt["metric"]

        return moval_model
    
# Dynamically remove the set_fit_request method
del MOVAL.set_fit_request


# modified from
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/cebra/registry.py
#
"""A simple registry for python modules.

This module only exposes a single public function, `add_helper_functions`,
which takes a python module or module name (or package) as its argument and
defines the decorator functions

* ``register``
* ``parametrize``

and the functions

* ``init``
* ``get_options``
* ``get_estim_options`` and 
* ``get_conf_options``

within this module. It also (implicitly and lazy) initializes a singleton
registry object which holds all registered classes. Typically, the helper
functions should be added in the first lines of a package ``__init__.py``
module.

Note that all functions carrying the respective decorators need to be discovered
by the import system, otherwise they will not be available when calling ``get_options``
or ``init``.
"""

from __future__ import annotations

import fnmatch
import itertools
import sys
import textwrap
import types
import warnings
from typing import Any, Dict, List, Union


class _Registry:
    _instance: Dict = None

    @classmethod
    def get_instance(cls, module: types.ModuleType) -> _Registry:
        if cls._instance is None:
            cls._instance = {}
        if module not in cls._instance:
            cls._instance[module] = {}
        return cls._instance[module]

    @classmethod
    def add(
        cls,
        module,
        name,
        value,
        base=None,
        override: bool = True,
        deprecated: bool = False,
    ):
        if not isinstance(name, str):
            raise ValueError(
                f"Name used for registration has to be str, got {type(name)}.")
        if not isinstance(value, type):
            raise ValueError(f"Can only register classes, got {type(value)}.")
        instance = cls.get_instance(module)
        if name in instance:
            if override:
                del instance[name]
            else:
                raise ValueError(f"Name {name} is already registered for "
                                 f"class {instance[name]}.")
        if (value, base) in instance.values():
            if not deprecated:
                raise ValueError(
                    f"Class {value} is already registered. "
                    f"Do you want to keep a deprecated alias for a module now "
                    f"registered under a new name? Then consider registering using "
                    f"@register(..., deprecated = True).")
        instance[name] = (value, base)

    @classmethod
    def init(cls, module, name, *args, **kwargs):
        instance = cls.get_instance(module)
        if name not in instance.keys():
            raise ValueError(f"name is {name}, not found in options")
        cls_, base = instance.get(name)
        return cls_(*args, **kwargs)

    @classmethod
    def get_options(
        cls,
        module,
        *,
        pattern: str = None,
        limit: int = None,
        expand_parametrized: bool = True,
    ):
        instance = cls.get_instance(module)
        if expand_parametrized:
            filter_ = lambda k, v: True
        else:

            class _Filter(set):

                def __call__(self, k, v):
                    if v is None:
                        return True
                    if v in self:
                        return False
                    self.add(v)
                    return True

            filter_ = _Filter()
        options = [k for k, (_, v) in instance.items() if filter_(k, v)]
        if pattern is None:
            return options[:limit]
        else:
            return fnmatch.filter(options, pattern)[:limit]


def _get_module(module: Union[types.ModuleType, str]) -> types.ModuleType:
    if isinstance(module, str):
        if module in sys.modules:
            return sys.modules[module]
        else:
            raise ValueError(
                f"Invalid module name: Cannot find module with name "
                f"{module} in sys.modules")
    if isinstance(module, types.ModuleType):
        return module
    raise TypeError(
        f"Invalid type: Expected str or module, but got {type(module)}")


def add_helper_functions(module: Union[types.ModuleType, str]):
    """Add registry functionality to the given module.

    Call this function within a python module to add the three
    functions ``register``, ``init``, ``get_options`, ``get_estim_options`` and ``get_conf_options`` to the
    module.

    * ``register`` is a decorator for classes within the module. Each class will be added by a (unique) name and can be initialized with the ``init`` function.
    * ``init`` takes a name as its argument and returns an instance of the specified class, with optional arguments.
    * ``get_options`` returns a list of all registered names within the module.
    * ``get_estim_options`` returns a list of all registered names within the base module of Model.
    * ``get_conf_options`` returns a list of all registered names within the base module of Confidence.

    Args:
        module: The module for adding registry functions. This can be
          the name of a module as returned by ``__name__`` within the
          module, or by passing the module type directory.

    """

    module = _get_module(module)

    def register(name: str,
                 base: str = None,
                 override: bool = False,
                 deprecated: bool = False):
        """Decorator to add a new class type to the registry."""

        def _register(cls):
            _Registry.add(module,
                          name,
                          cls,
                          base=base,
                          override=override,
                          deprecated=deprecated)
            return cls

        return _register

    def parametrize(pattern: str,
                    *,
                    kwargs: List[Dict[str, Any]] = [],
                    **all_kwargs):
        """Decorator to add parametrizations of a new class to the registry.

        The specified keyword arguments will be passed as default arguments
        to the constructor of the class.
        """

        def _product_dict(d):
            keys = tuple(d.keys())
            values = tuple(d[k] for k in keys)
            combinations = itertools.product(*values)
            for combination in combinations:
                yield dict(zip(keys, combination))

        def _zip_dict(d):
            keys = tuple(d.keys())
            values = tuple(d[k] for k in keys)
            combinations = itertools.product(*values)
            for combination in combinations:
                yield dict(zip(keys, combination))

        def _create_class(cls, **default_kwargs):

            @register(pattern.format(**default_kwargs), base=pattern)
            class _ParametrizedClass(cls):

                def __init__(self, *args, **kwargs):
                    default_kwargs.update(kwargs)
                    super().__init__(*args, **default_kwargs)

        def _parametrize(cls):
            for _default_kwargs in kwargs:
                _create_class(cls, **_default_kwargs)
            if len(all_kwargs) > 0:
                for _default_kwargs in _product_dict(all_kwargs):
                    _create_class(cls, **_default_kwargs)
            return cls

        return _parametrize

    def init(name: str, *args, **kwargs):
        """Initialize an instance from the registry with the specified arguments.

        Args:
            name: The to identify the registered class
            args, kwargs: Arguments and keyword arguments to pass to the
                constructor while instantiating the selected type.

        Returns:
            An instance of the specified class.
        """
        return _Registry.init(module, name, *args, **kwargs)

    def get_options(pattern: str = None,
                    limit: int = None,
                    expand_parametrized: bool = True) -> List[str]:
        """Retrieve a list of registered names, optionally filtered.

        Args:
            pattern: A glob-like pattern (supporting wildcards ``*`` and ``?`` to
                filter the options. Optional argument, defaults to no filtering.
            limit: An optional maximum amount of options to return, in the order
                of finding them with the given query.
            expand_parametrized: Whether to list classes registered with the
                ``parametrize`` decorator in the options.

        Returns:
            All matching names. If a ``limit`` was specified, the maximum length
            is given by the limit.
        """
        return _Registry.get_options(
            module,
            pattern=pattern,
            limit=limit,
            expand_parametrized=expand_parametrized,
        )

    def get_estim_options(pattern: str = None,
                    limit: int = None,
                    expand_parametrized: bool = True) -> List[str]:
        """Retrieve a list of registered names, optionally filtered.

        Args:
            pattern: A glob-like pattern (supporting wildcards ``*`` and ``?`` to
                filter the options. Optional argument, defaults to no filtering.
            limit: An optional maximum amount of options to return, in the order
                of finding them with the given query.
            expand_parametrized: Whether to list classes registered with the
                ``parametrize`` decorator in the options.

        Returns:
            All matching names. If a ``limit`` was specified, the maximum length
            is given by the limit.
        """
        return _Registry.get_options(
            module,
            pattern="*-model",
            limit=limit,
            expand_parametrized=expand_parametrized,
        )

    def get_conf_options(pattern: str = None,
                    limit: int = None,
                    expand_parametrized: bool = True) -> List[str]:
        """Retrieve a list of registered names, optionally filtered.

        Args:
            pattern: A glob-like pattern (supporting wildcards ``*`` and ``?`` to
                filter the options. Optional argument, defaults to no filtering.
            limit: An optional maximum amount of options to return, in the order
                of finding them with the given query.
            expand_parametrized: Whether to list classes registered with the
                ``parametrize`` decorator in the options.

        Returns:
            All matching names. If a ``limit`` was specified, the maximum length
            is given by the limit.
        """
        return _Registry.get_options(
            module,
            pattern="*-conf",
            limit=limit,
            expand_parametrized=expand_parametrized,
        )

    names = ["register", "init", "get_options", "get_estim_options", "get_conf_options", "parametrize"]
    for name in names:
        if hasattr(module, name):
            raise RuntimeError(
                f"Specified module {module.__name__} already defines {module.__name__}.{name}. "
                "Overriding existing functions is not possible. Make sure that "
                "add_helper_functions is only called once, and that the function names "
                f"{names} are not previously defined in the module.")

    module.register = register
    module.parametrize = parametrize
    module.init = init
    module.get_options = get_options
    module.get_estim_options = get_estim_options
    module.get_conf_options = get_conf_options

    if not is_registry(module):
        raise RuntimeError(
            f"Registry could not be successfully registered: {module}.")


def add_docstring(module: Union[types.ModuleType, str]):
    """Apply additional information about configuration options to registry modules.

    Args:
        module: Name of the module, or the module itself. If a string is
            given, it needs to match the representation in ``sys.modules``.
    """

    def _shorten(text):
        return textwrap.shorten(str(text),
                                width=80,
                                break_on_hyphens=False,
                                placeholder=" ...]")

    def _wrap(text, indent: int):
        return textwrap.fill(str(text),
                             subsequent_indent=" " * 4,
                             break_on_hyphens=False)

    module = _get_module(module)

    options = module.get_options(limit=10)
    toplevel_name = module.__name__.split(".")[-1]

    if len(options) < 1:
        warnings.warn(
            f"Called {__name__}.add_docstring inside the module {module.__name__} which does not register",
            "any classes. Did you import submodules using the registration decorator?",
            ImportWarning,
        )

    if not is_registry(module):
        raise ImportError(
            f"Cannot call {__name__}.add_docstring on module {module.__name__} which did"
            f"not previously call {module.__name__}.add_helper_functions.")

    docstring = f"""\
    This module is a registry and currently contains the options
    {_wrap(options, 4)}.

    To retrieve a list of options, call::

        >>> print({module.__name__}.get_options())
        {_shorten(options)}

    To obtain an initialized instance, call ``{module.__name__}.init``,
    defined in :py:func:`moval.registry.add_helper_functions`.
    The first parameter to provide is the {toplevel_name} name to use,
    which is one of the available options presented above.
    Then the required positional arguments specific to the module are provided, if
    needed.

    You can register additional options by defining and registering
    classes with a name. To do that, you can add a decorator on top of it:
    ``@{module.__name__}.register("my-{module.__name__.replace('.', '-')}")``.

    Later, initialize your class similarly to the pre-defined options, using ``{module.__name__}.init``
    with the {toplevel_name} name set to ``my-{module.__name__.replace('.', '-')}``.

    Note that these customized options will not be automatically added to this
    docstring.
    """
    docstring = textwrap.dedent(docstring)

    module.__doc__ = "\n\n".join([module.__doc__, docstring])


def is_registry(module: Union[types.ModuleType, str],
                check_docs: bool = False) -> bool:
    """Check if the given module implements all registry functions.

    Args:
        module: Name of the module, or the module itself. If a string is
            given, it needs to match the representation in ``sys.modules``.
        check_docs: Optionally specify whether or not to check if a docstring
            was adapted, specifying all default options.

    Returns:
        True if the module is a registry and implements the ``register``, ``init``
        , ``get_options``, ``get_estim_options`` and ``get_conf_options`` functions. If ``check_docs`` is set to ``True``, then
        the documentation needs to match in addition. False if at least one function
        is missing.
    """

    module = _get_module(module)
    if check_docs:
        for option in module.get_options(limit=10):
            if option not in module.__doc__:
                return False
    return all(
        hasattr(module, name)
        for name in ["register", "parametrize", "init", "get_options", "get_estim_options", "get_conf_options"])



#
# Adapted from:
# https://github.com/ZerojumpLine/ModelEvaluationUnderClassImbalance
#
"""Variants of solvers to train model performance estimation algorithms.
    
This packages contain optimiziers to align the probability and accuracy.

"""

import moval.registry

moval.registry.add_helper_functions(__name__)

from moval.solvers.solver import *
from moval.solvers.criterions import *

moval.registry.add_docstring(__name__)


import abc
from typing import Callable, Iterable, List, Literal, Optional, Tuple, Union
import numpy as np
from moval.solvers import register
from moval.models.utils import cal_softmax
from moval.solvers.utils import ComputMetric, ComputAUC

class Calibrate(abc.ABC):
    """Base model for calibration criterions

    Features can be computed by calling the ``__call__()`` method. This class should not 
    be directly instaniated, and instead used as the base class for calibration criterions.

    """

    def __call__(
        self,
        inp: Union[List[Iterable], np.ndarray],
        gt: Union[List[Iterable], np.ndarray],
        estim: np.ndarray
    ) -> np.ndarray:
        """Compute the calibration error.
        """
        raise NotImplementedError()

@register("cls-calibrate")
class clsCalibrate(Calibrate):
    def __init__(self, class_specific=False, metric="accuracy"):
        super().__init__()
        self.class_specific = class_specific
        self.metric = metric
    
    def __call__(
        self,
        inp: np.ndarray,
        gt: np.ndarray,
        estim: List[Iterable],
        kcls_all: List[Iterable] = None
    ) -> np.ndarray:
        """Compute the calibration err for classification tasks.
        
        Args:
            inp: The network output (logits) of shape ``(n, d)``. 
            gt: The cooresponding annotation of shape ``(n, )``.
            estim: The estimated performance by moval, this is a float for class agnostic version.
                For class-specific version, estim is list of shape ``(1, )`` which contains estimation for kcls. 
            kcls_all: This contains the class list, only for the class-specific verision, which indicates the class to be optimized.
        
        Returns:
            err: The difference between the estimated accuracy and the real accuracy.
                For class-specific version, err contains differences for class kcls.
        
        """

        pred = np.argmax(inp, axis = 1)

        if not self.class_specific:
            
            acc = np.sum(gt == pred) / len(gt)
            err = estim - acc

        else:
            w_kcls_all = []
            for kcls in kcls_all:
                pos_cls = np.where(pred == kcls)[0]
                w_kcls_all.append(len(pos_cls))        
            w_kcls_all = w_kcls_all / np.sum(w_kcls_all) # normalize
            err = 0
            k_cnt = 0
            if self.metric == "accuracy":
                e1 = 1e-6 # small number to avoid nan.
                for kcls in kcls_all:
                    pos_cls = np.where(pred == kcls)[0]
                    acc_cls = np.sum(gt[pos_cls] == pred[pos_cls]) / (len(gt[pos_cls]) + e1)
                    err += w_kcls_all[k_cnt] * (estim[k_cnt] - acc_cls)
                    k_cnt += 1

            elif self.metric == "sensitivity":
                for kcls in kcls_all:
                    _, sensitivity_cls, _ = ComputMetric(gt == kcls, pred == kcls)
                    err += w_kcls_all[k_cnt] * (estim[k_cnt] - sensitivity_cls)
                    k_cnt += 1

            elif self.metric == "precision":
                for kcls in kcls_all:
                    _, _, precision_cls = ComputMetric(gt == kcls, pred == kcls)
                    err += w_kcls_all[k_cnt] * (estim[k_cnt] - precision_cls)
                    k_cnt += 1

            elif self.metric == "f1score":
                for kcls in kcls_all:
                    f1score_cls, _, _ = ComputMetric(gt == kcls, pred == kcls)
                    err += w_kcls_all[k_cnt] * (estim[k_cnt] - f1score_cls)
                    k_cnt += 1

            elif self.metric == "auc":
                for kcls in kcls_all:
                    aucs = ComputAUC(gt, cal_softmax(inp), kcls)
                    err += w_kcls_all[k_cnt] * (estim[k_cnt] - aucs)
                    k_cnt += 1

            else:
                ValueError(f"Unsupported metric '{self.metric}'")

        return np.abs(err)

@register("seg-calibrate")
class segCalibrate(Calibrate):
    def __init__(self, class_specific=False, metric="accuracy"):
        super().__init__()
        self.class_specific = class_specific
        self.metric = metric
    
    def __call__(
        self,
        inp: List[Iterable],
        gt: List[Iterable],
        estim: List[Iterable],
        kcls_all: List[Iterable] = None
    ) -> np.ndarray:
        """Compute the calibration err for segmentation tasks.
        
        Args:
            inp: The network output (logits) as a list of n ``(d, H, W, (D))`` for segmentation. 
            gt: The cooresponding annotation as a list of n ``(H, W, (D))`` for segmentation. 
            estim: The estimated performance by moval, this is a float for class agnostic version.
                For class-specific version, estim is list of shape ``(1, )`` which contains estimation for kcls. 
            kcls_all: This contains the class list, only for the class-specific verision, which indicates the class to be optimized.
        
        Returns:
            err: The difference between the estimated metric and the real metric.
                For class-specific version, err contains differences for class kcls.

        """

        if not self.class_specific:
            pred_all_flatten = []
            gt_all_flatten = []
            for n_case in range(len(inp)):

                pred_case   = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                gt_case     = gt[n_case] # ``(H, W, (D))``
                
                pred_all_flatten.append(pred_case.flatten())
                gt_all_flatten.append(gt_case.flatten())
            
            pred_all_flatten = np.concatenate(pred_all_flatten)
            gt_all_flatten = np.concatenate(gt_all_flatten)

            acc = np.sum(gt_all_flatten == pred_all_flatten) / len(gt_all_flatten)
            err = estim - acc

            return np.abs(err)

        else:
            estim_kcls = estim[0]
            kcls = kcls_all[0]

            if self.metric == "accuracy":
                
                err = np.zeros(inp[0].shape[0])
                    
                pred_all_flatten_kcls = []
                gt_all_flatten_kcls = []

                for n_case in range(len(inp)):

                    pred_case   = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                    gt_case     = gt[n_case] # ``(H, W, (D))``
                    pos_kcls = np.where(pred_case.flatten() == kcls)[0]
                    pred_all_flatten_kcls.append(pred_case.flatten()[pos_kcls])
                    gt_all_flatten_kcls.append(gt_case.flatten()[pos_kcls])
                
                pred_all_flatten_kcls = np.concatenate(pred_all_flatten_kcls, axis=0)
                gt_all_flatten_kcls = np.concatenate(gt_all_flatten_kcls, axis=0)

                acc_cls = np.sum(gt_all_flatten_kcls == pred_all_flatten_kcls) / len(gt_all_flatten_kcls)
                err = estim_kcls - acc_cls
                
            elif self.metric == "sensitivity":

                sensitivity = []
                for n_case in range(len(inp)):

                    pred_case   = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                    gt_case     = gt[n_case] # ``(H, W, (D))``

                    if np.sum(gt_case == kcls) == 0:
                        sensitivity_cls = -1
                    else:
                        _, sensitivity_cls, _ = ComputMetric(gt_case == kcls, pred_case == kcls)
                    sensitivity.append(sensitivity_cls)
                
                # only aggregate the ones which are not -1
                sensitivity = np.array(sensitivity) # ``(n, )``
                # I am not sure, if the real sensitivity is 0, I think the network cannot learn anything
                # but any calculate this case in the estim, may we can improve here.
                sensitivity_mean = sensitivity[sensitivity >= 0].mean()

                err = estim_kcls - sensitivity_mean

            elif self.metric == "precision":

                precision = []
                for n_case in range(len(inp)):

                    pred_case   = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                    gt_case     = gt[n_case] # ``(H, W, (D))``

                    if np.sum(gt_case == kcls) == 0:
                        precision_cls = -1
                    else:
                        _, _, precision_cls = ComputMetric(gt_case == kcls, pred_case == kcls)
                    precision.append(precision_cls)
                
                # only aggregate the ones which are not -1
                precision = np.array(precision) # ``(n, d)``
                # I am not sure, if the real precision is 0, I think the network cannot learn anything
                # but any calculate this case in the estim, may we can improve here.
                precision_mean = precision[precision >= 0].mean()

                err = estim_kcls - precision_mean

            elif self.metric == "f1score":

                dsc = []
                for n_case in range(len(inp)):

                    pred_case   = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                    gt_case     = gt[n_case] # ``(H, W, (D))``

                    if np.sum(gt_case == kcls) == 0:
                        dsc_cls = -1
                    else:
                        dsc_cls, _, _ = ComputMetric(gt_case == kcls, pred_case == kcls)
                    dsc.append(dsc_cls)
                
                # only aggregate the ones which are not -1
                dsc = np.array(dsc) # ``(n, d)``
                # I am not sure, if the real dsc is 0, I think the network cannot learn anything
                # but any calculate this case in the estim, may we can improve here.
                dsc_mean = dsc[dsc >= 0].mean()

                err = estim_kcls - dsc_mean

            elif self.metric == "auc":
                
                auc = []
                for n_case in range(len(inp)):
                    
                    inp_case = inp[n_case] # ``(d, H, W, (D))``
                    # from ``(d, H, W, (D))`` to ``(n, d)``
                    d, *rest_of_dimensions = inp_case.shape
                    flatten_dim = np.prod(rest_of_dimensions)
                    inp_case = inp_case.reshape((d, flatten_dim))
                    inp_case = inp_case.T # ``(n, d)``
                    probability = cal_softmax(inp_case) # ``(n, d)``                    
                    gt_case     = gt[n_case].flatten() # ``(n, )``

                    auc_case = ComputAUC(gt_case, probability, kcls) # ``(1, )``
                    auc.append(auc_case)
                
                # only aggregate the ones which are not -1
                auc = np.array(auc) # ``(n, 1)``
                # I am not sure, if the real dsc is 0, I think the network cannot learn anything
                # but any calculate this case in the estim, may we can improve here.
                auc_mean = auc[:, 0][auc[:, 0] >= 0].mean()

                err = estim_kcls - auc_mean

            else:
                ValueError(f"Unsupported metric '{self.metric}'")

            return np.abs(err)


"""Scikit-Learn interface for MOVAL

The implementation follows the guide on `Developing scikit-learn estimators`_.

Note:
    The main class of this module, :py:class:`MOVAL`, is also available under the
    top-level package name as :py:class:`moval.MOVAL` and automatically imported
    with :py:mod:`moval`.

.. _Developing scikit-learn estimators:
    https://scikit-learn.org/stable/developers/develop.html
"""

from moval.integrations.sklearn import moval


# Adapted from
# https://github.com/melanibe/failure_detection_benchmark

import abc
from typing import Callable, Iterable, List, Literal, Optional, Tuple, Union

from moval.models import register
from moval.models.utils import cal_energy, cal_mcp, cal_entropy, cal_doctor
import numpy as np

class Confidence(abc.ABC):
    """Base model for confidence calculation.

    Confidence scores can be computed by calling the ``__call__()`` method. This class should not 
    be directly instaniated, and instead used as the base class for confidence scores calculation.

    Atrributes:
        score_type (str):
            The name of confidence scores.
        negative_score (bool):
            A boolean that indicates if original confidence score is negative correlated to the classification correctness.
        normalization (bool):
            A boolean that indicates if original confidence score needs normalization before usage.
        cal_func (:py:func):
            The function to calculate the corresponding confidence scores.

    """

    def __init__(self, score_type: str, negative_score: bool, normalization: bool):
        self.score_type = score_type
        self.negative_score = negative_score
        self.normalization = normalization
        
        if self.score_type == "max_class_probability-conf":
            self.cal_func = cal_mcp
        elif self.score_type == "energy-conf":
            self.cal_func = cal_energy
        elif self.score_type == "entropy-conf":
            self.cal_func = cal_entropy
        elif self.score_type == "doctor-conf":
            self.cal_func = cal_doctor
        else:
            raise ValueError(f"Unknown mode '{self.mode}'")
    
    def reshape_inp(self, inp: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Reshape the input from arbitrary shape to ``(n, d)``.

        Args:
            inp: The network output (logits) of shape ``(n, d, H, W, (D))``.
        
        Returns:
            inp: Reshaped network output of shape ``(n, d)``.
            inp_shape: The shape of original network output.
        
        """
        
        inp_shape = inp.shape

        # flaten the logit for segmentation.
        # first tranpose the feature to the last dimension
        axes = [0] + list(range(2, len(inp_shape))) + [1]
        inp = np.transpose(inp, axes)
        # then flatten all
        inp = inp.reshape((-1,) + inp.shape[-1:])
        
        return inp, inp_shape
    
    def __call__(self, inp: Union[List[Iterable], np.ndarray], T: np.ndarray = None) -> Union[List[Iterable], np.ndarray]:
        """Calculate softmax probability and return the maximum.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            T: Tempratature to calibrate the confidence scores of shape ``(d, )`` for class-specific or ``(1, )`` for class-agnostic.
                If it is omited as None, the scores are not calibrated (T = 1).
        
        Return:
            conf_score: The calculated confidence scores, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
        
        """
        if T is None:
            T = np.array([1])

        if isinstance(inp, list):
            # inp is a list, segmentaiton task
            num_class = inp[0].shape[0]

            conf_scores = []

            for n_case in range(len(inp)):

                inp_case = inp[n_case] # (d, H, W, (D))
                inp_case = inp_case[np.newaxis, ...] # (1, d, H, W, (D))

                inp_case, inp_shape = self.reshape_inp(inp_case)
                # now inp_case is of shape ``(n, d)``

                pred = np.argmax(inp_case, axis=1) # ``(n, )``
                conf_score = np.zeros(inp_case.shape[0]) # ``(n, )``
                
                if len(T) == 1:
                    conf_score = self.cal_func(inp_case, T)
                else:
                    for kcls in range(num_class):
                        pred_cls_position,  = np.where(pred == kcls)
                        conf_score[pred_cls_position] = self.cal_func(inp_case[pred_cls_position, :], T[kcls])

                if self.negative_score:
                    conf_score = 1 - conf_score

                # transfer back to the original dimension
                conf_score = conf_score.reshape(inp_shape[2:]) # ``(H, W, (D))``
            
                conf_scores.append(conf_score)

            return conf_scores

        else:
            # inp is not a list, classification task

            num_class = inp.shape[1]

            pred = np.argmax(inp, axis=1)
            conf_score = np.zeros(inp.shape[0]) # ``(n, )``
            
            if len(T) == 1:
                conf_score = self.cal_func(inp, T)
            else:
                for kcls in range(num_class):
                    pred_cls_position,  = np.where(pred == kcls)
                    conf_score[pred_cls_position] = self.cal_func(inp[pred_cls_position, :], T[kcls])

            if self.negative_score:
                conf_score = 1 - conf_score

            return conf_score

@register("max_class_probability-conf")
class maxclassprobabilityConfidence(Confidence):
    """Confidence scores based on max class probability.
    
    Note:
        The maximum class probability ([0, 1]) is high is the model is confident.
    
    """

    def __init__(self):
        super().__init__(
            score_type = "max_class_probability-conf",
            negative_score = False,
            normalization = False
        )
    
    
# energy scores
@register("energy-conf")
class energyConfidence(Confidence):
    """Confidence scores based on energy.
    
    Note:
        Energy is low is the model is confident, we should calculate negative energy.
        Original engergy is unbounded, we should normalize by finding the max/min value later.

    """

    def __init__(self):
        super().__init__(
            score_type = "energy-conf",
            negative_score = True,
            normalization = True
        )

# entropy
@register("entropy-conf")
class entropyConfidence(Confidence):
    """Confidence scores based on entropy.
    
    Note:
        Entropy is low is the model is confident, we should calculate negative entropy.
        We should also choose the base as class number to ensure that the largest entropy is 1.

    """

    def __init__(self):
       super().__init__(
            score_type = "entropy-conf",
            negative_score = True,
            normalization = False
        )
    

# doctor scores
@register("doctor-conf")
class doctorConfidence(Confidence):
    """MOVAL model with doctor confidence."""

    def __init__(self):
        super().__init__(
            score_type = "doctor-conf",
            negative_score = True,
            normalization = True
        )
    


#
# Adapted from:
# https://github.com/ZerojumpLine/ModelEvaluationUnderClassImbalance
#
"""Variants of model performance estimation algorithms.

This package contains model definitions of performance estimation and calculation of different confidence scores.

"""

import moval.registry

moval.registry.add_helper_functions(__name__)

from moval.models.model import *
from moval.models.confidences import *

moval.registry.add_docstring(__name__)




"""MOVAL is tool to estimate model performance without manual labels. 
It contains several confidences score calculation and calibration algorithms, implemented with python and sklearn.
It can estimate accuracy for classification tasks and dice scores for segmentation tasks.
"""

try:
    from moval.integrations.sklearn.moval import MOVAL
except ImportError as e:
    # silently fail for now
    pass

__version__ = "0.3.25"
__all__ = ["MOVAL"]

def __getattr__(key):
    """Lazy import of moval submodules and -packages.

    Once :py:mod:`moval` is imported, it is possible to lazy import

    """
    if key == "MOVAL":
        from moval.integrations.sklearn.moval import MOVAL

        return MOVAL


from sklearn.metrics import auc, roc_curve
from moval.models.utils import one_hot_embedding
import numpy as np

def ComputMetric(ACTUAL: np.ndarray, PREDICTED: np.ndarray) -> float:
    """Calculate the DSC (F1-Score), sensitivity and precision between prediction and the GT.

    Args:
        PREDICTED: The predicted segmentation map of shape ``(H, W, (D))`` or predicted classification results of shape ``(n, )``.
        ACTUAL: The ground truth segmentation map of shape ``(H, W, (D))`` or predicted classification results of shape ``(n, )``.

    Returns:
        dice: A float scalar which represents the calculated dice score.
        sensitivity: A float scalar which represents the calculated sensitivity score.
        precision: A float scalar which represents the calculated precision score.
    
    """
    ACTUAL = ACTUAL.flatten()
    PREDICTED = PREDICTED.flatten()
    idxp = ACTUAL == True
    idxn = ACTUAL == False

    tp = np.sum(ACTUAL[idxp] == PREDICTED[idxp])
    tn = np.sum(ACTUAL[idxn] == PREDICTED[idxn])
    fp = np.sum(idxn) - tn
    fn = np.sum(idxp) - tp
    if tp == 0 :
        dice = 0
        sensitivity = 0
        precision = 0
    else:
        dice = 2 * tp / (2 * tp + fp + fn)
        precision = tp / (tp + fp)
        sensitivity = tp / (tp + fn)
    return dice, sensitivity, precision

# AUC calculation here.
def ComputAUC(ACTUAL: np.ndarray, PROBABILITY: np.ndarray, sel_cls: int = None) -> float:
    """Calculate the AUC.

    Note:
        When dealing with multi-class AUC, we calculate the One-vs-Rest strategy for simplicity.
        More information could refer to https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html.
        It is not used in the framework of MOVAL, but could be useful to validate moval's performance.

    Args:
        ACTUAL: The ground truth segmentation map of shape ``(H, W, (D))`` or predicted classification results of shape ``(n, )``.
        PROBABILITY: The predicted probability of shape ``(d, H, W, (D))`` or predicted classification results of shape ``(n, d)``.
        sel_cls: The selected class for calculation. If it is None, return all classes.
    
    Returns:
        AUCs: The calculate class-wise AUC of shape ``(d, )``, or ``(1, )`` if sel_cls is givien.
    
    """

    ACTUAL = ACTUAL.flatten() # ``(n, )``
    if len(PROBABILITY.shape) > 2:
        # from ``(d, H, W, (D))`` to ``(n, d)``
        d, *rest_of_dimensions = PROBABILITY.shape
        flatten_dim = np.prod(rest_of_dimensions)
        PROBABILITY = PROBABILITY.reshape((d, flatten_dim))
        PROBABILITY = PROBABILITY.T

    y_onehot_test = one_hot_embedding(ACTUAL, PROBABILITY.shape[1])
    fpr, tpr, roc_auc = dict(), dict(), dict()

    if sel_cls is None:
        allcls = range(PROBABILITY.shape[1])
    else:
        allcls = range(sel_cls, sel_cls+1)

    for i in allcls:
        
        fpr[i], tpr[i], _ = roc_curve(y_onehot_test[:, i], PROBABILITY[:, i])
        roc_auc[i] = auc(fpr[i], tpr[i])

    return np.array(list(roc_auc.values()))




import abc
from typing import Callable, Iterable, List, Literal, Optional, Tuple, Union
import numpy as np
from moval.models import register
import moval.models
from moval.models.utils import SoftDiceLoss, SoftSensitivity, SoftPrecision, SoftAUC, cal_softmax
from moval.models.solver_temperature import solve_T

class Model(abc.ABC):
    """Base model for MOVAL experiments.

    Features can be computed by calling the ``__call__()`` method. This class should not 
    be directly instaniated, and instead used as the base class for MOVAl models.

    Args:
        mode (str): The given task to estimate model performance.
        num_class (int): The number of class for the given task.
        confidence_scores (str): The method to calculate confidence scores.
        class_specific (bool): If ``True``, the calculation will match class-wise confidence to class-wise accuracy/DSC.
    
    Atrributes:
        estim_algorithm (str):
            The chosen estimated algorithms for confidence calibration.
        mode (str):
            The given task to estimate model performance.
            This will choose the criterions for parameter optimization.
        num_class (int): 
            The number of class for the given task.
            This will decide the number of parameter to optimize and optimization criterions.
        confidence_scores (str): 
            The method to calculate confidence scores. 
            This will decide the methods chosen from confidences.
        class_spcific (bool):
            If ``True``, the calculation will match class-wise confidence to class-wise accuracy/DSC.
            This will also affect the number of parameter to optimize and optimization criterions.
        conf (moval.models.Confidence):
            The Confidence class to calculate the confidence scores.
        max_value (float):
            The maximum value to normalize the confidence score before calibration.
            This is only necessary when :py:attr:`conf` is unbounded.
        min_value (float):
            The minimum value to normalize the confidence score before calibration.
            This is only necessary when :py:attr:`conf` is unbounded.
        param (np.ndarray):
            The parameter to be optimized.
        extend_param (bool):
            If ``True``, the model contains more parameters (x2) and need two-stage optimization.
            Currently, this is only applicable for ``ts-atc-model``.
        is_fitted (bool):
            If ``True``, the model is fitted using validation data and ready to use for estimating performance.
        is_training (bool):
            If ``True``, the normalization paramters would be updated using the input data.

    """

    def __init__(
        self,
        estim_algorithm: str,
        mode: str,
        num_class: int,
        confidence_scores: str,
        class_specific: bool,
        extend_param: bool
    ):
        self.estim_algorithm = estim_algorithm
        self.mode = mode
        self.num_class = num_class
        self.confidence_scores = confidence_scores
        self.class_specific = class_specific
        self.is_fitted = False
        self.extend_param = extend_param

        self.conf = moval.models.init(self.confidence_scores)
        
        self.is_training = False
        self.normalized = False
        
        # Note, for non-normalized method, we choose the parameter as the lower bound for normalization.
        if self.conf.normalization:
            self.max_value = 0.
            self.min_value = 0.

        if self.class_specific:
            self.param = np.ones(num_class)
        else:
            self.param = np.array([1])
        
        # allocate addtional set of parameters.
        if self.extend_param:
            if self.class_specific:
                self.param_ext = np.ones(num_class)
            else:
                self.param_ext = np.array([1])
    
    def train(self):
        """Initialize the model into training mode.
        """
        self.is_training = True
    
    def eval(self):
        """Initialize the model into evaluation mode.
        """
        self.is_training = False

    def _find_normalize(self, scores: Union[List[Iterable], np.ndarray]):
        """Find the normalization parameter for the confidence scores.

        Args:
            scores: The unbounded confidence scores, of shape ``(n, )`` or a list of n ``(H, W, (D))``.

        Notes:
            The intensity value, max and min value, will be saved. This should be only done with the T = 1 (inital states).
        
        """
        if isinstance(scores, list):
            max_value_list = []
            min_value_list = []
            for score in scores:
                max_value_list.append(np.max(score))
                min_value_list.append(np.min(score))
            self.max_value = np.max(max_value_list)
            self.min_value = np.min(min_value_list)
        
        else:
            self.max_value = np.max(scores)
            self.min_value = np.min(scores)

    def _normalize(self, scores: Union[List[Iterable], np.ndarray], e1: float = 1e-6) -> Union[List[Iterable], np.ndarray]:
        """Normalize the confidence scores to [0, 1] if the score is not bounded.

        Note:
            This replace the temperature scaling process of this confidence scores, e.g., doctor, energy.

        Args:
            scores: The unbounded confidence scores, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
            e1: A small number to prevent unexpected results of division.
        
        Returns:
            normalized_scores: The normalized confidence scores, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
        
        """

        if self.is_training and not self.normalized:
            
            self._find_normalize(scores)
            self.normalized = True

        if isinstance(scores, list):
            
            normalized_scores = []
            for n_case in range(len(scores)):
                normalized_score =  (scores[n_case] - self.min_value) / (self.max_value - self.min_value + e1)
                # normalized_score has shape ``(H, W, (D))``
                # restrict the scores to the range of [0, 1]
                normalized_score = np.clip(normalized_score, 0, 1)
                normalized_scores.append(normalized_score)
        else:

            normalized_scores = (scores - self.min_value) / (self.max_value - self.min_value + e1)
            # normalized_scores has shape ``(n, )``
            # restrict the scores to the range of [0, 1]
            normalized_scores = np.clip(normalized_scores, 0, 1)
                
        return normalized_scores
    
    def estimate_accuracy(self, inp: Union[List[Iterable], np.ndarray], midstage: bool = False, gt_guide: np.ndarray = None) -> Tuple[float, np.ndarray]:
        """Estimate the accuracy using network output.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            midstage: If ``True``, return the first calibrated results.
            gt_guide: The cooresponding annotation guide of shape ``(n, d)`` for segmentation. This is only rquired for segmentation task during optimizaing.
                We will utilize this to determine if there is label in the case. If not, we do not calculate the dsc and utilize for optimizing.
                This is because we do not want to optimize parameter with blank segmentation map. 
                This should be bool, if ``False``, it means that there isn't any manuel label of class d in this sample.
        
        Returns:
            estim_acc: A float scalar that represents the estimated accuracy for the given input data.
            estim_cls: Estimated class-wise accuracy of shape ``(d, )``.

        """

        # calibrate the confidence scores
        if self.extend_param:
            score = self.calibrate(inp, midstage)
        else:
            score = self.calibrate(inp)

        # Estimate the performance.
        if self.mode == "classification":
            # for classification tasks, return average confidence
            estim_acc = np.mean(score)
            # class-wise average confidence
            pred = np.argmax(inp, axis = 1)
            estim_acc_allcls = []
            for kcls in range(inp.shape[1]):
                score_kcls = score[np.where(pred == kcls)[0]]
                if len(score_kcls) > 0:
                    estim_acc_kcls = np.mean(score_kcls)
                else:
                    # does not pred any
                    estim_acc_kcls = 1e-5
                estim_acc_allcls.append(estim_acc_kcls)
            return estim_acc, np.array(estim_acc_allcls)
        elif self.mode == "segmentation":
            # for segmentation tasks
            scores = []
            for n_case in range(len(inp)):
                score_flatten = score[n_case].flatten() # ``n``
                scores.append(score_flatten)
            estim_acc = np.mean( np.concatenate(scores) )

            estim_acc_allcls = []
            for kcls in range(inp[0].shape[0]):
                scores_kcls = []
                for n_case in range(len(inp)):
                    pred_case = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                    pred_flatten = pred_case.flatten() # ``n``
                    score_flatten = score[n_case].flatten() # ``n``
                    score_flatten_kcls = score_flatten[np.where(pred_flatten == kcls)[0]]
                    scores_kcls.append(score_flatten_kcls)
                estim_acc_allcls.append(np.mean( np.concatenate(scores_kcls) ))
            
            return estim_acc, np.array(estim_acc_allcls)
        else:
            raise ValueError(f"Unknown mode '{self.mode}'")

    def estimate_sensitivity(self, inp: Union[List[Iterable], np.ndarray], probability: Union[List[Iterable], np.ndarray], gt_guide: np.ndarray = None) -> Tuple[float, np.ndarray]:
        """Esimate the sensitivity using network output.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            probability: The calibrated probability of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            gt_guide: The cooresponding annotation guide of shape ``(n, d)`` for segmentation. This is only rquired for segmentation task during optimizaing.
                We will utilize this to determine if there is label in the case. If not, we do not calculate the dsc and utilize for optimizing.
                This is because we do not want to optimize parameter with blank segmentation map. 
                This should be bool, if ``False``, it means that there isn't any manuel label of class d in this sample.
        
        Note:
            The user may wonder why we need inp here in this function. It is because we utilize inp to determine the prediction results, which are not always feasible by probability.

        Returns:
            estim_sensitivity: Estimated class-wise sensitivity of shape ``(d, )``.

        """

        # Estimate the sensitivity.
        if self.mode == "classification":
            # probability is of shape ``(n, d)``
            estim_sensitivity = SoftSensitivity(probability, np.argmax(inp, axis = 1))
            return estim_sensitivity
        elif self.mode == "segmentation":
            # probability is a list of n ``(d, H, W, (D))``.
            estim_sensitivity_list = []
            for n_case in range(len(inp)):
                pred_case = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                score_filled = probability[n_case] # ``(d, H, W, (D))``
                #
                estim_sensitivity = SoftSensitivity(score_filled[np.newaxis, ...], pred_case[np.newaxis, ...])
                #
                if isinstance(gt_guide, np.ndarray):
                    gt_case = gt_guide[n_case]
                    # We remove the class DSC if there isn't any in gt
                    for kcls in range(len(estim_sensitivity)):
                        if gt_case[kcls] is False:
                            estim_sensitivity[kcls] = -1

                estim_sensitivity_list.append(estim_sensitivity)
            
            # aggregate the results
            estim_sensitivity_list = np.array(estim_sensitivity_list) # ``(n, d)``
            m_estim_sensitivity = []
            for kcls in range(len(estim_sensitivity)):
                m_estim_sensitivity.append(estim_sensitivity_list[:, kcls][estim_sensitivity_list[:,kcls] >= 0].mean())
            
            return np.array(m_estim_sensitivity)
        else:
            raise ValueError(f"Unknown mode '{self.mode}'")
    
    def estimate_precision(self, probability: Union[List[Iterable], np.ndarray], gt_guide: np.ndarray = None) -> Tuple[float, np.ndarray]:
        """Esimate the precision using network output.

        Args:
            probability: The calibrated probability of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            gt_guide: The cooresponding annotation guide of shape ``(n, d)`` for segmentation. This is only rquired for segmentation task during optimizaing.
                We will utilize this to determine if there is label in the case. If not, we do not calculate the dsc and utilize for optimizing.
                This is because we do not want to optimize parameter with blank segmentation map. 
                This should be bool, if ``False``, it means that there isn't any manuel label of class d in this sample.
        
        Returns:
            estim_precision: Estimated class-wise precision of shape ``(d, )``.

        """

        if self.estim_algorithm == "atc-model" or self.estim_algorithm == "ts-atc-model":
            raise ValueError(f"estimation algorithm '{self.estim_algorithm}' does not support the estimation of precision (because FP is always 0)")

        # Estimate the sensitivity.
        if self.mode == "classification":
            # probability is of shape ``(n, d)``
            estim_precision = SoftPrecision(probability, np.argmax(probability, axis = 1))
            return estim_precision
        elif self.mode == "segmentation":
            # probability is a list of n ``(d, H, W, (D))``.
            estim_precision_list = []
            for n_case in range(len(probability)):
                pred_case = np.argmax(probability[n_case], axis = 0) # ``(H, W, (D))``
                score_filled = probability[n_case] # ``(d, H, W, (D))``
                #
                estim_precision = SoftPrecision(score_filled[np.newaxis, ...], pred_case[np.newaxis, ...])
                #
                if isinstance(gt_guide, np.ndarray):
                    gt_case = gt_guide[n_case]
                    # We remove the class DSC if there isn't any in gt
                    for kcls in range(len(estim_precision)):
                        if gt_case[kcls] is False:
                            estim_precision[kcls] = -1

                estim_precision_list.append(estim_precision)
            
            # aggregate the results
            estim_precision_list = np.array(estim_precision_list) # ``(n, d)``
            m_estim_precision = []
            for kcls in range(len(estim_precision)):
                m_estim_precision.append(estim_precision_list[:, kcls][estim_precision_list[:,kcls] >= 0].mean())
            
            return np.array(m_estim_precision)
        else:
            raise ValueError(f"Unknown mode '{self.mode}'")
    
    def estimate_f1score(self, inp: Union[List[Iterable], np.ndarray], probability: Union[List[Iterable], np.ndarray], gt_guide: np.ndarray = None) -> Tuple[float, np.ndarray]:
        """Esimate the F1score using network output.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            probability: The calibrated probability of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            gt_guide: The cooresponding annotation guide of shape ``(n, d)`` for segmentation. This is only rquired for segmentation task during optimizaing.
                We will utilize this to determine if there is label in the case. If not, we do not calculate the dsc and utilize for optimizing.
                This is because we do not want to optimize parameter with blank segmentation map. 
                This should be bool, if ``False``, it means that there isn't any manuel label of class d in this sample.
        
        Returns:
            estim_F1score: Estimated class-wise F1score of shape ``(d, )``.

        """

        # Estimate the sensitivity.
        if self.mode == "classification":
            # probability is of shape ``(n, d)``
            estim_F1score = SoftDiceLoss(probability, np.argmax(inp, axis = 1))
            return estim_F1score
        elif self.mode == "segmentation":
            # probability is a list of n ``(d, H, W, (D))``.
            estim_dsc_list = []
            for n_case in range(len(inp)):
                pred_case = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                score_filled = probability[n_case] # ``(d, H, W, (D))``
                #
                estim_dsc = SoftDiceLoss(score_filled[np.newaxis, ...], pred_case[np.newaxis, ...])
                #
                if isinstance(gt_guide, np.ndarray):
                    gt_case = gt_guide[n_case]
                    # We remove the class DSC if there isn't any in gt
                    for kcls in range(len(estim_dsc)):
                        if gt_case[kcls] is False:
                            estim_dsc[kcls] = -1

                estim_dsc_list.append(estim_dsc)
            
            # aggregate the results
            estim_dsc_list = np.array(estim_dsc_list) # ``(n, d)``
            m_estim_dsc = []
            for kcls in range(len(estim_dsc)):
                m_estim_dsc.append(estim_dsc_list[:, kcls][estim_dsc_list[:,kcls] >= 0].mean())
            
            return np.array(m_estim_dsc)
        else:
            raise ValueError(f"Unknown mode '{self.mode}'")
    
    def estimate_auc(self, 
                     probability: Union[List[Iterable], np.ndarray], 
                     gt_guide: np.ndarray = None, 
                     sel_cls: int = None) -> Tuple[float, np.ndarray]:
        """Esimate the AUC using network output.

        Args:
            probability: The calibrated probability of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            gt_guide: The cooresponding annotation guide of shape ``(n, d)`` for segmentation. This is only rquired for segmentation task during optimizaing.
                We will utilize this to determine if there is label in the case. If not, we do not calculate the dsc and utilize for optimizing.
                This is because we do not want to optimize parameter with blank segmentation map. 
                This should be bool, if ``False``, it means that there isn't any manuel label of class d in this sample.
            sel_cls: The selected class for calculation. If it is None, return all classes.
        
        Returns:
            estim_AUC: Estimated class-wise AUC of shape ``(d, )``, or ``(1, )`` if sel_cls is givien.

        """

        if self.estim_algorithm == "atc-model" or self.estim_algorithm == "ts-atc-model":
            raise ValueError(f"estimation algorithm '{self.estim_algorithm}' does not support the estimation of AUC (because FP is always 0)")

        # Estimate the sensitivity.
        if self.mode == "classification":
            # probability is of shape ``(n, d)``
            estim_AUC, _, _ = SoftAUC(probability, sel_cls = sel_cls)
            return estim_AUC
        elif self.mode == "segmentation":
            # probability is a list of n ``(d, H, W, (D))``.
            estim_AUC_list = []
            for n_case in range(len(probability)):
                score_filled = probability[n_case] # ``(d, H, W, (D))``
                #
                estim_AUC, _, _ = SoftAUC(score_filled[np.newaxis, ...], sel_cls = sel_cls)
                #
                if isinstance(gt_guide, np.ndarray):
                    gt_case = gt_guide[n_case]
                    # We remove the class DSC if there isn't any in gt
                    for kcls in range(len(estim_AUC)):
                        if gt_case[kcls] is False:
                            estim_AUC[kcls] = -1

                estim_AUC_list.append(estim_AUC)
            
            # aggregate the results
            estim_AUC_list = np.array(estim_AUC_list) # ``(n, d)``
            m_estim_auc = []
            for kcls in range(len(estim_AUC)):
                m_estim_auc.append(estim_AUC_list[:, kcls][estim_AUC_list[:,kcls] >= 0].mean())
            
            return np.array(m_estim_auc)
        else:
            raise ValueError(f"Unknown mode '{self.mode}'")

    def calibrate(self, inp: Union[List[Iterable], np.ndarray]) -> Union[List[Iterable], np.ndarray]:
        """Calibrate the confidence scores with parameters.

        Different estimation algorithms would adopt different strateiges to calibrate the scores.
        
        """
        
        raise NotImplementedError()
    
    def calculate_probability(self, inp: Union[List[Iterable], np.ndarray], midstage: bool = False, appr: bool = False, full: bool = False) -> Union[List[Iterable], np.ndarray]:
        """Calculate the calibrated probability with parameters.
        For classificaiton tasks, we choose to estimate the pseudo-temperature, for segmentation tasks, we simplify it with 1-socre.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            midstage: If ``True``, return the first calibrated results.
            appr: If ``True``, utilize the approximation version.
            full: If ``False``, we just put the score in and fill the other domains with 0.

        Returns:
            calibrated_probability: The calibrated probability which would match the accuracy/DSC on validation data, of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
        
        """

        if self.mode == "segmentation" or appr:
            return self.calculate_probability_appr(inp, midstage, full)
        elif self.mode == "classification":
            return self.calculate_probability_temperature(inp, midstage, full)
        else:
            raise ValueError(f"Unknown mode '{self.mode}'")

    def calculate_probability_temperature(self, inp: Union[List[Iterable], np.ndarray], midstage: bool = False, full : bool = False) -> Union[List[Iterable], np.ndarray]:
        """Calculate the calibrated probability with parameters, based on temperature scaling.
        The challenge is the calculation of non-maximum probability. 
        To achieve this, we calculate the pseudo-temperature for each samples such that the confidence score match the max probability after the temperature scaling process.
        Then, we utilize the pseudo-temperature to scale the probabilities of other classes.

        Note:
            The calibrated probability should have the same dimension with network outputs.
            The calculation is slow as we need to go through every samples.
            Also, it might be not accurate when calculating estimation algorithms other than MCP, as they are in range [0, 1] instead of [1/n, 1].
        
        Update:
            The function is barely used now, as it is too slow!

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            midstage: If ``True``, return the first calibrated results.
            full: If ``False``, we just put the score in and fill the other domains with 0.

        Returns:
            calibrated_probability: The calibrated probability which would match the accuracy/DSC on validation data, of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
        
        """

        if self.extend_param:
            score = self.calibrate(inp, midstage)
        else:
            score = self.calibrate(inp)
            
        if self.mode == "classification":
            # extend from ``(n, )`` to ``(n, d)``
            if not full or self.estim_algorithm == "atc-model" or (self.estim_algorithm == "ts-atc-model" and midstage == False):
                # ATC cannot get the calibrated probability of all classes, we just fill the other dimension with zeros
                pred_flatten = np.argmax(inp, axis = 1)
                probability = np.zeros((score.shape + (self.num_class,))) # ``(n, d)``
                probability[np.arange(probability.shape[0]), pred_flatten] = score # ``(n, d)``
            else:
                T = solve_T(inp, score) # return T of shape ``(n, )``

                probability = np.zeros((score.shape + (self.num_class,))) # ``(n, d)``
                for ksample in range(len(score)):
                    probability[ksample, :] = cal_softmax(inp[ksample: ksample + 1], T[ksample]) # ``(1, d)``

        elif self.mode == "segmentation":
            # extend from a list of n ``(H, W, (D))`` to a list of n ``(d, H, W, (D))``
            if not full or self.estim_algorithm == "atc-model" or (self.estim_algorithm == "ts-atc-model" and midstage == False):
                # ATC cannot get the calibrated probability of all classes, we just fill the other dimension with zeros
                probability = []
                for n_case in range(len(inp)):
                    pred_case = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                    #
                    pred_flatten = pred_case.flatten() # ``n``
                    score_case = score[n_case] # ``(H, W, (D))``
                    score_flatten = score[n_case].flatten() # ``n``
                    probability_case = np.zeros((score_flatten.shape + (self.num_class,))) # ``(n, d)``
                    probability_case[np.arange(probability_case.shape[0]), pred_flatten] = score_flatten # ``(n, d)``
                    probability_case = probability_case.T # ``(d, n)``
                    probability_case = probability_case.reshape(((self.num_class,) + score_case.shape)) # ``(d, H, W, (D))``
                    probability.append(probability_case)
            else:
                probability = []
                for n_case in range(len(inp)):
                    score_case = score[n_case] # ``(H, W, (D))``
                    score_flatten = score[n_case].flatten() # ``n``
                    inp_case = inp[n_case] # ``(d, H, W, (D))``
                    inp_case = inp_case.reshape(((self.num_class, len(score_flatten)))) # ``(d, n)``
                    inp_case = inp_case.T # ``(n, d)``
                    T = solve_T(inp_case, score_flatten) # return T of shape ``(n, )``
                    probability_case = np.zeros((score_flatten.shape + (self.num_class,))) # ``(n, d)``
                    for ksample in range(len(score)):
                        probability_case[ksample, :] = cal_softmax(inp_case[ksample: ksample + 1], T[ksample]) # ``(1, d)``
                    probability_case = probability_case.T # ``(d, n)``
                    probability_case = probability_case.reshape(((self.num_class,) + score_case.shape)) # ``(d, H, W, (D))``
                    probability.append(probability_case)

        else:
            raise ValueError(f"Unknown mode '{self.mode}'")
        
        return probability
    
    def calculate_probability_appr(self, inp: Union[List[Iterable], np.ndarray], midstage: bool = False, full : bool = False) -> Union[List[Iterable], np.ndarray]:
        """Calculate the calibrated probability with parameters, utilizing approximation of 1-score.
        To acclerate the optimization process, we calculate the probability but divide 1-score equally to other classes.
        I should use it for all the segmentation tasks, as the pixel number is always quite large.

        Note:
            The calibrated probability should have the same dimension with network outputs.
            This is identical to calculate_probability when d == 2, i.e. binary segmentation, which is the most common case. Therefore, I am not very worry about the descrepencies.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            midstage: If ``True``, return the first calibrated results.
            full: If ``False``, we just put the score in and fill the other domains with 0.

        Returns:
            calibrated_probability: The calibrated probability which would match the accuracy/DSC on validation data, of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
        
        """

        if self.extend_param:
            score = self.calibrate(inp, midstage)
        else:
            score = self.calibrate(inp)
            
        if self.mode == "classification":
            # extend from ``(n, )`` to ``(n, d)``
            pred_flatten = np.argmax(inp, axis = 1)
            probability = np.zeros((score.shape + (self.num_class,))) # ``(n, d)``
            if not full or self.estim_algorithm == "atc-model" or (self.estim_algorithm == "ts-atc-model" and midstage == False):
                # ATC cannot get the calibrated probability of all classes, we just fill the other dimension with zeros
                pass
            else:
                # first fill the probability with score / (d-1)
                score_extended = np.tile(score, (self.num_class,1)).transpose()
                probability[np.arange(probability.shape[0]), :] = (1 - score_extended) / (self.num_class - 1) # ``(n, d)``
            
            probability[np.arange(probability.shape[0]), pred_flatten] = score # overwrite the confidence score of the predicted class

        elif self.mode == "segmentation":
            # extend from a list of n ``(H, W, (D))`` to a list of n ``(d, H, W, (D))``
            
            probability = []
            for n_case in range(len(inp)):
                pred_case = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                #
                pred_flatten = pred_case.flatten() # ``n``
                score_case = score[n_case] # ``(H, W, (D))``
                score_flatten = score[n_case].flatten() # ``n``
                probability_case = np.zeros((score_flatten.shape + (self.num_class,))) # ``(n, d)``
                if not full or self.estim_algorithm == "atc-model" or (self.estim_algorithm == "ts-atc-model" and midstage == False):
                    # ATC cannot get the calibrated probability of all classes, we just fill the other dimension with zeros
                    pass
                else:
                    # first fill the probability with score / (d-1)
                    score_extended = np.tile(score_flatten, (self.num_class,1)).transpose()
                    probability_case[np.arange(probability_case.shape[0]), :] = (1 - score_extended) / (self.num_class - 1) # ``(n, d)``

                probability_case[np.arange(probability_case.shape[0]), pred_flatten] = score_flatten # ``(n, d)``
                probability_case = probability_case.T # ``(d, n)``
                probability_case = probability_case.reshape(((self.num_class,) + score_case.shape)) # ``(d, H, W, (D))``
                probability.append(probability_case)

        else:
            raise ValueError(f"Unknown mode '{self.mode}'")
        
        return probability

@register("ac-model")
class acModel(Model):
    """MOVAL model with average confidence."""

    def __init__(self, mode: str, num_class: int, confidence_scores: str, class_specific: bool):
        super().__init__(
            estim_algorithm = "ac-model",
            mode = mode,
            num_class = num_class,
            confidence_scores = confidence_scores,
            class_specific = class_specific,
            extend_param = False
        )
    
    def calibrate(self, inp: Union[List[Iterable], np.ndarray]) -> Union[List[Iterable], np.ndarray]:
        """Calibrate the confidence scores for average confidence.
        
        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.

        Returns:
            calibrated_scores: The calibrated scores which would match the accuracy/DSC on validation data, of shape ``(n, )`` or a list of n ``(H, W, (D))``.

        """

        # Calualte the confidence socres.
        score = self.conf(inp) # ``(n,)`` for classification or a list of n ``(H, W, (D))`` for segmentation

        # Normalize the scores, if needed.
        if self.conf.normalization:
            score = self._normalize(score)
        
        return score

@register("ts-model")
class tsModel(Model):
    """MOVAL model with temperature scaling."""

    def __init__(self, mode: str, num_class: int, confidence_scores: str, class_specific: bool):
        super().__init__(
            estim_algorithm = "ts-model",
            mode = mode,
            num_class = num_class,
            confidence_scores = confidence_scores,
            class_specific = class_specific,
            extend_param = False
        )
    
    def calibrate(self, inp: Union[List[Iterable], np.ndarray]) -> Union[List[Iterable], np.ndarray]:
        """Calibrate the confidence scores with temperature scaling.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.

        Returns:
            calibrated_scores: The calibrated scores which would match the accuracy/DSC on validation data, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
        
        """

        # Calualte the confidence socres.
        score = self.conf(inp, self.param) # ``(n,)`` for classification or a list of n ``(H, W, (D))`` for segmentation

        # Normalize the scores, if needed.
        if self.conf.normalization:
            score = self._normalize(score)
        
        return score

@register("doc-model")
class docModel(Model):
    """MOVAL model with difference of confidence."""

    def __init__(self, mode: str, num_class: int, confidence_scores: str, class_specific: bool):
        super().__init__(
            estim_algorithm = "doc-model",
            mode = mode,
            num_class = num_class,
            confidence_scores = confidence_scores,
            class_specific = class_specific,
            extend_param = False
        )
    
    def calibrate(self, inp: Union[List[Iterable], np.ndarray]) -> Union[List[Iterable], np.ndarray]:
        """Calibrate the confidence scores based on difference of confidence.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.

        Returns:
            calibrated_scores: The calibrated scores which would match the accuracy/DSC on validation data, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
        
        Note:
            The implementation of class-specific DoC for segmentation is different from what we did in our MICCAI work.
            In MICCAI, we calculate the difference as the difference of soft dsc. Here, we modify the confidence score to match dsc.
            Current version should be more compatable with other techiques.
        
        """

        # Calualte the confidence socres.
        score = self.conf(inp) # ``(n,)`` for classification or a list of n ``(H, W, (D))`` for segmentation

        # Normalize the scores, if needed.
        if self.conf.normalization:
            score = self._normalize(score)

        if not self.class_specific:
            if self.mode == "classification":
                score = score - (1 - self.param)
                return np.clip(score, 0, 1)
            elif self.mode == "segmentation":
                score_post = []
                for n_case in range(len(score)):
                    score_case = score[n_case] - (1 - self.param)
                    score_post.append(np.clip(score_case, 0, 1))
                return score_post
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")
        else:
            # class-wise average confidence
            if self.mode == "classification":
                pred = np.argmax(inp, axis = 1)
                for kcls in range(inp.shape[1]):
                    pos_cls = np.where(pred == kcls)[0]
                    score[pos_cls] = score[pos_cls] - (1 - self.param[kcls])
                return np.clip(score, 0, 1)
            elif self.mode == "segmentation":
                score_post = []
                for n_case in range(len(score)):
                    score_case = score[n_case] # ``(H, W, (D))``
                    pred_case = np.argmax(inp[n_case], axis = 0)
                    score_shape = score_case.shape
                    score_flatten = score_case.flatten()
                    pred_flatten = pred_case.flatten()
                    #
                    for kcls in range(self.num_class):
                        pos_flatten_cls = np.where(pred_flatten == kcls)[0]
                        score_flatten[pos_flatten_cls] = score_flatten[pos_flatten_cls] - (1 - self.param[kcls])
                        score_reshape = score_flatten.reshape(score_shape) # ``(H, W, (D))``
                    score_post.append(np.clip(score_reshape, 0, 1))
                return score_post
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")

@register("atc-model")
class atcModel(Model):
    """MOVAL model with average thresholded confidence."""

    def __init__(self, mode: str, num_class: int, confidence_scores: str, class_specific: bool):
        super().__init__(
            estim_algorithm = "atc-model",
            mode = mode,
            num_class = num_class,
            confidence_scores = confidence_scores,
            class_specific = class_specific,
            extend_param = False
        )

    def calibrate(self, inp: Union[List[Iterable], np.ndarray]) -> Union[List[Iterable], np.ndarray]:
        """Calibrate the confidence scores based on average thresholded confidence.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.

        Returns:
            calibrated_scores: The calibrated scores which would match the accuracy/DSC on validation data, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
        
        """

        # Calualte the confidence socres.
        _score = self.conf(inp) # ``(n,)`` for classification or a list of n ``(H, W, (D))`` for segmentation

        # Normalize the scores, if needed.
        if self.conf.normalization:
            _score = self._normalize(_score)

        if not self.class_specific:
            if self.mode == "classification":
                score = np.zeros(inp.shape[0])
                score[_score > self.param] = 1
                return score
            elif self.mode == "segmentation":
                score_post = []
                for n_case in range(len(_score)):
                    score = np.zeros(_score[n_case].shape)
                    score[_score[n_case] > self.param] = 1
                    score_post.append(score)
                return score_post
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")
        else:
            # class-wise average confidence
            if self.mode == "classification":
                score = np.zeros(inp.shape[0])
                pred = np.argmax(inp, axis = 1)
                for kcls in range(inp.shape[1]):
                    pos_cls = np.where(pred == kcls)[0]
                    score_pos = score[pos_cls]
                    score_pos[_score[pos_cls] > self.param[kcls]] = 1
                    score[pos_cls] = score_pos
                return score
            elif self.mode == "segmentation":
                score_post = []
                for n_case in range(len(_score)):
                    _score_case = _score[n_case] # ``(H, W, (D))``
                    score_case = np.zeros(_score[n_case].shape)
                    pred_case = np.argmax(inp[n_case], axis = 0)
                    score_shape = score_case.shape
                    pred_flatten = pred_case.flatten()
                    score_flatten = score_case.flatten()
                    _score_flatten = _score_case.flatten()
                    #
                    for kcls in range(self.num_class):
                        pos_flatten_cls = np.where(pred_flatten == kcls)[0]
                        score_pos = score_flatten[pos_flatten_cls]
                        score_pos[_score_flatten[pos_flatten_cls] > self.param[kcls]] = 1
                        score_flatten[pos_flatten_cls] = score_pos
                        score_reshape = score_flatten.reshape(score_shape) # ``(H, W, (D))``
                    score_post.append(score_reshape)
                return score_post
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")

@register("ts-atc-model")
class tsatcModel(Model):
    """MOVAL model with average thresholded confidence after temperature scaling."""

    def __init__(self, mode: str, num_class: int, confidence_scores: str, class_specific: bool):
        super().__init__(
            estim_algorithm = "ts-atc-model",
            mode = mode,
            num_class = num_class,
            confidence_scores = confidence_scores,
            class_specific = class_specific,
            extend_param = True
        )
    
    def calibrate(self, inp: Union[List[Iterable], np.ndarray], midstage: bool = False) -> Union[List[Iterable], np.ndarray]:
        """Calibrate the confidence scores based on average thresholded confidence after temperature scaling.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            midstage: If ``True``, return the first calibrated results.

        Returns:
            calibrated_scores: The calibrated scores which would match the accuracy/DSC on validation data, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
        
        """

        # Calualte the confidence socres.
        _score = self.conf(inp, self.param) # ``(n,)`` for classification or a list of n ``(H, W, (D))`` for segmentation
        
        # Normalize the scores, if needed.
        if self.conf.normalization:
            _score = self._normalize(_score)

        if midstage:
            return _score

        if not self.class_specific:
            if self.mode == "classification":
                score = np.zeros(inp.shape[0])
                score[_score > self.param_ext] = 1
                return score
            elif self.mode == "segmentation":
                score_post = []
                for n_case in range(len(_score)):
                    score = np.zeros(_score[n_case].shape)
                    score[_score[n_case] > self.param_ext] = 1
                    score_post.append(score)
                return score_post
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")
        else:
            # class-wise average confidence
            if self.mode == "classification":
                score = np.zeros(inp.shape[0])
                pred = np.argmax(inp, axis = 1)
                for kcls in range(inp.shape[1]):
                    pos_cls = np.where(pred == kcls)[0]
                    score_pos = score[pos_cls]
                    score_pos[_score[pos_cls] > self.param_ext[kcls]] = 1
                    score[pos_cls] = score_pos
                return score
            elif self.mode == "segmentation":
                score_post = []
                for n_case in range(len(_score)):
                    _score_case = _score[n_case] # ``(H, W, (D))``
                    score_case = np.zeros(_score[n_case].shape)
                    pred_case = np.argmax(inp[n_case], axis = 0)
                    score_shape = score_case.shape
                    pred_flatten = pred_case.flatten()
                    score_flatten = score_case.flatten()
                    _score_flatten = _score_case.flatten()
                    #
                    for kcls in range(self.num_class):
                        pos_flatten_cls = np.where(pred_flatten == kcls)[0]
                        score_pos = score_flatten[pos_flatten_cls]
                        score_pos[_score_flatten[pos_flatten_cls] > self.param_ext[kcls]] = 1
                        score_flatten[pos_flatten_cls] = score_pos
                        score_reshape = score_flatten.reshape(score_shape) # ``(H, W, (D))``
                    score_post.append(score_reshape)
                return score_post
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")